import multiprocessing as mp
import os
import platform
import shutil
import sys
import timeit
from functools import partial as func_partial
from pathlib import PosixPath

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray
from alphabase.io.tempmmap import create_empty_mmap, mmap_array_from_path
from anndata import AnnData
from scipy.ndimage import binary_fill_holes
from skimage.filters import gaussian
from spatialdata import SpatialData
from tqdm.auto import tqdm

from scportrait.pipeline._base import ProcessingStep
from scportrait.pipeline._utils.helper import flatten
from scportrait.processing.images._image_processing import percentile_normalization


class HDF5CellExtraction(ProcessingStep):
    """
    A class to extracts single cell images from a segmented scPortrait project and save the
    results to an HDF5 file.
    """

    CLEAN_LOG = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

        self._check_config()

        # setup base extraction workflow
        self._get_compression_type()

        # initialize base variables to save results to
        self.save_index_to_remove = []
        self.batch_size = None

        # set developer debug mode for super detailed output
        self.deep_debug = False

        # check for windows operating system and if so set threads to 1
        if platform.system() == "Windows":
            Warning("Windows detected. Multithreading not supported on windows so setting threads to 1.")
            self.threads = 1

        if "overwrite_run_path" not in self.__dict__.keys():
            self.overwrite_run_path = self.overwrite

        self.extraction_file = os.path.join(self.directory, self.DEFAULT_DATA_DIR, self.DEFAULT_EXTRACTION_FILE)
        self.output_path = None

    def _get_compression_type(self) -> None:
        """setup compression of single-cell images in HDF5 file based on config."""
        # default value for compression is "lzf" is nothing else is specified
        if (self.compression is True) or (self.compression == "lzf"):
            self.compression_type = "lzf"
        elif self.compression == "gzip":
            self.compression_type = "gzip"
        else:
            self.compression_type = None
        self.log(f"Compression algorithm for extracted single-cell images: {self.compression_type}")

    def _check_config(self):
        """Load parameters from config file and check for required parameters."""

        # check for required parameters
        assert "threads" in self.config, "Number of threads must be specified in the config file."
        assert "image_size" in self.config, "Image size must be specified in the config file."
        assert "cache" in self.config, "Cache directory must be specified in the config file."

        self.threads = self.config["threads"]
        self.image_size = self.config["image_size"]

        # optional parameters with default values that can be overridden by the values in the config file

        ## parameters for image extraction
        if "normalize_output" in self.config:
            normalize = self.config["normalize_output"]

            # check that is a valid value
            assert normalize in [
                True,
                False,
                None,
                "None",
            ], "Normalization must be one of the following values [True, False, None, 'None']"

            # convert to boolean
            if normalize == "None":
                normalize = False
            if normalize is None:
                normalize = False

            self.normalization = normalize

        else:
            self.normalization = True  # default value

        if "normalization_range" in self.config:
            normalization_range = self.config["normalization_range"]

            if normalization_range == "None":
                normalization_range = None

            if normalization_range is not None:
                assert len(normalization_range) == 2, "Normalization range must be a tuple or list of length 2."
                assert all(
                    isinstance(x, float | int) and (0 <= x <= 1) for x in normalization_range
                ), "Normalization range must be defined as a float between 0 and 1."

                # conver to tuple to ensure consistency
                if isinstance(normalization_range, list):
                    normalization_range = tuple(normalization_range)
            else:
                normalization_range = (0.01, 0.99)

            self.normalization_range = normalization_range

        else:
            self.normalization_range = [0.01, 0.99]

        if not self.normalization:
            self.normalization_range = ("None", "None")

        ## parameters for HDF5 file creates
        if "compression" in self.config:
            self.compression = self.config["compression"]
        else:
            self.compression = True

        ## Deprecated parameters since we no longer directly read from HDF5
        ## Preservering here in case we see better performance by adjusting this behaviour in how we read data from memmapped arrays

        # if "hdf5_rdcc_nbytes" in self.config:
        #     self.hdf5_rdcc_nbytes = self.config["hdf5_rdcc_nbytes"]
        # else:
        #     self.hdf5_rdcc_nbytes = 5242880000 # 5gb 1024 * 1024 * 5000

        # if "hdf5_rdcc_w0" in self.config:
        #     self.hdf5_rdcc_w0 = self.config["hdf5_rdcc_w0"]
        # else:
        #     self.hdf5_rdcc_w0 = 1

        # if "hdf5_rdcc_nslots" in self.config:
        #     self.hdf5_rdcc_nslots = self.config["hdf5_rdcc_nslots"]
        # else:
        #     self.hdf5_rdcc_nslots = 50000

    def _setup_normalization(self) -> None:
        """Configure normalization of single-cell images based on config.

        Default behaviour is that images are 1-99 percentile normalized and converted to a float between 0 and 1.

        This default behaviour can be overridden by providing a boolean value to normalize_output to configure if the images should be normalized, or
        by providing a tuple to normalization_range to specify the lower and upper percentile values to use for normalization.

        If normalization is set to False, the images will be converted to a float between 0 and 1 without any normalization where the highest value is the maximum value of the image datatype.
        """
        if self.normalization:
            lower, upper = self.normalization_range

            def percentile_norm(img: np.ndarray, lower: float = lower, upper: float = upper) -> np.ndarray:
                return percentile_normalization(img, lower, upper)

            self.norm_function = percentile_norm

        else:

            def min_max(img: np.ndarray, lower=None, upper=None) -> np.ndarray:
                img = (
                    img / np.iinfo(self.DEFAULT_IMAGE_DTYPE).max
                )  # convert 16bit unsigned integer image to float between 0 and 1 without adjusting for the pixel values we have in the extracted single cell image
                return img

            self.norm_function = min_max

    def _get_output_path(self) -> str | PosixPath:
        """Get the output path for the extraction results."""
        return self.extraction_data_directory

    def _setup_output(self, folder_name: str | None = None) -> None:
        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)

        # if a foldername is not specified it uses the default value
        if folder_name is None:
            folder_name = self.DEFAULT_DATA_DIR

        self.extraction_data_directory = os.path.join(self.directory, folder_name)

        if not os.path.isdir(self.extraction_data_directory):
            os.makedirs(self.extraction_data_directory)
            self.log(f"Created new directory for extraction results: {self.extraction_data_directory}")
        else:
            if self.overwrite_run_path:
                self.log(f"Output folder at {self.extraction_data_directory} already exists. Overwriting...")
                shutil.rmtree(self.extraction_data_directory)
                os.makedirs(self.extraction_data_directory)
                self.log(f"Created new directory for extraction results: {self.extraction_data_directory}")

            elif self.overwrite_run_path:
                self.log(f"Output folder at {self.extraction_data_directory} already exists. Overwriting...")
                shutil.rmtree(self.extraction_data_directory)
                os.makedirs(self.extraction_data_directory)
                self.log(f"Created new directory for extraction results: {self.extraction_data_directory}")

            else:
                raise ValueError("Output folder already exists. Set overwrite_run_path to True to overwrite.")

        self.log(f"Setup output folder at {self.extraction_data_directory}")

    def _set_up_extraction(self, output_folder_name: str | None) -> None:
        """execute all helper functions to setup extraction process"""

        if output_folder_name is None:
            output_folder_name = self.DEFAULT_DATA_DIR

        if self.partial_processing:
            output_folder_name = f"partial_{output_folder_name}_ncells_{self.n_cells}_seed_{self.seed}"
        else:
            output_folder_name = output_folder_name

        self._setup_output(folder_name=output_folder_name)
        self._get_segmentation_info()
        self._get_input_image_info()

        # setup number of output channels
        self.n_output_channels = self.n_image_channels + self.n_masks

        # get size of images to extract
        self.extracted_image_size = self.image_size
        self.width_extraction = (
            self.extracted_image_size // 2
        )  # half of the extracted image size (this is what needs to be added on either side of the center)

        self._get_centers()
        self._get_classes_to_extract()

        # create output files for saving results to
        self._create_output_files()

        # print relevant information to log file
        self._verbalise_extraction_info()

    def _get_segmentation_info(self) -> None:
        """execute logic to determine which segmentation masks to use for extraction"""

        # get current version of spatialdata object
        _sdata = self.filehandler._read_sdata()

        segmentation_keys = list(_sdata.labels.keys())

        # specific segmentation masks can be extracted that do not need to conform to the default naming convention in scportrait
        # this behaviour can be achieved by passing a segmentation_key in the config file.
        if "segmentation_key" in self.config.keys():
            self.segmentation_key = self.config["segmentation_key"]
        else:
            self.segmentation_key = "seg_all"  # default segmentation key

        relevant_masks = [x for x in segmentation_keys if self.segmentation_key in x]

        # ensure that we have at least 1 mask to proceed with
        if len(relevant_masks) == 0:
            raise ValueError(
                f"Found no segmentation masks with key {self.segmentation_key}. Cannot proceed with extraction."
            )

        # intialize default values to track what should be extracted
        self.nucleus_key = None
        self.cytosol_key = None
        self.extract_nucleus_mask = False
        self.extract_cytosol_mask = False

        if "segmentation_mask" in self.config:
            if isinstance(self.config["segmentation_mask"], str):
                if "nucleus" in self.config["segmentation_mask"]:
                    self.nucleus_key = self.config["segmentation_mask"]
                    self.extract_nucleus_mask = True
                elif "cytosol" in self.config["segmentation_mask"]:
                    self.cytosol_key = self.config["segmentation_mask"]
                    self.extract_cytosol_mask = True
                else:
                    Warning(
                        "Unclear if the singular provided segmentation mask is a nucleus or cytosol mask. Will proceed assuming cytosol like behaviour."
                    )
                    self.cytosol_key = self.config["segmentation_mask"]
                    self.extract_cytosol_mask = True

            elif isinstance(self.config["segmentation_mask"], list):
                for x in self.config["segmentation_mask"]:
                    if "nucleus" in x:
                        self.nucleus_key = x
                        self.extract_nucleus_mask = True
                    elif "cytosol" in x:
                        self.cytosol_key = x
                        self.extract_cytosol_mask = True
                    else:
                        raise ValueError(
                            f"Segmentation mask {self.config['segmentation_mask']} does not indicate which mask is the nucleus and which is the cytosol. Cannot proceed with extraction."
                        )

        else:
            # get relevant segmentation masks to perform extraction on
            nucleus_key = f"{self.segmentation_key}_nucleus"

            if nucleus_key in relevant_masks:
                self.extract_nucleus_mask = True
                self.nucleus_key = nucleus_key

            cytosol_key = f"{self.segmentation_key}_cytosol"

            if cytosol_key in relevant_masks:
                self.extract_cytosol_mask = True
                self.cytosol_key = cytosol_key

        self.n_masks = np.sum([self.extract_nucleus_mask, self.extract_cytosol_mask])
        self.masks = [x for x in [self.nucleus_key, self.cytosol_key] if x is not None]

        # define the main segmentation mask to extract single-cell images from
        # this mask will be used to calculate the cell centers
        if self.n_masks == 2:
            # perform sanity check that the masks have the same ids
            # THIS NEEDS TO BE IMPLEMENTED HERE

            self.main_segmentation_mask = self.nucleus_key

        elif self.n_masks == 1:
            if self.extract_nucleus_mask:
                self.main_segmentation_mask = self.nucleus_key
            elif self.extract_cytosol_mask:
                self.main_segmentation_mask = self.cytosol_key

        self.log(
            f"Found {self.n_masks} segmentation masks for the given key in the sdata object. Will be extracting single-cell images based on these masks: {self.masks}"
        )
        self.log(f"Using {self.main_segmentation_mask} as the main segmentation mask to determine cell centers.")

    def _get_input_image_info(self) -> None:
        """get relevant information about the input image to be able to extract single-cell images"""
        # get channel information
        input_image = self.filehandler._get_input_image(self.filehandler.get_sdata())
        self.channel_names = input_image.c.values
        self.n_image_channels = len(self.channel_names)
        self.input_image_width = len(input_image.x)
        self.input_image_height = len(input_image.y)

    def _get_centers(self) -> None:
        """get the centers of the cells that should be extracted.
        If a nucleus and cytosol mask are used, the centers are calculated based on the nucleus mask.
        If only one mask is used, the centers are calculated based on that mask.
        """
        _sdata = self.filehandler._read_sdata()

        # calculate centers if they have not been calculated yet
        centers_name = f"{self.DEFAULT_CENTERS_NAME}_{self.main_segmentation_mask}"
        if centers_name not in _sdata:
            self.filehandler._add_centers(self.main_segmentation_mask, overwrite=self.overwrite)
            _sdata = self.filehandler._read_sdata()  # reread to ensure we have updated version

        centers = _sdata[centers_name].values.compute()

        # round to int so that we can use them as indices
        centers = np.round(centers).astype(int)

        self.centers = centers
        self.centers_cell_ids = _sdata[centers_name].index.values.compute()

        # ensure that the centers ids are unique
        assert len(self.centers_cell_ids) == len(
            set(self.centers_cell_ids)
        ), "Cell ids in centers are not unique. Cannot proceed with extraction."

        # double check that the cell_ids contained in the seg masks match to those from centers
        # THIS NEEDS TO BE IMPLEMENTED HERE

    def _get_classes_to_extract(self):
        """Get final list of unique cell IDs that should be extracted.
        Cell IDs for cells which are too close to the image edges to extract are discarded.
        """
        if self.partial_processing:
            self.log(
                f"Partial extraction mode enabled. Randomly sampling {self.n_cells} cells to extract with seed {self.seed}."
            )

            # randomly sample n_cells from the centers
            rng = np.random.default_rng(self.seed)
            chosen_ids = rng.choice(list(range(len(self.centers_cell_ids))), self.n_cells, replace=False)

            self.classes = self.centers_cell_ids[chosen_ids]
            self.px_centers = self.centers[chosen_ids]
        else:
            self.classes = self.centers_cell_ids
            self.px_centers = self.centers

        # ensure that none of the classes are located too close to the image edges to extract
        ids_to_discard = self._check_location_of_cells_to_extract()
        self._save_removed_classes(ids_to_discard)  # write discarded ids to file

        # update classes and centers to only contain the classes that should be extracted
        indexes_keep = [i for i, x in enumerate(self.classes) if x not in ids_to_discard]
        self.classes = self.classes[indexes_keep]
        self.px_centers = self.px_centers[indexes_keep]
        self.num_classes = len(self.classes)  # get total number of IDs that need to be processed

    def _check_location_of_cells_to_extract(self) -> list[int]:
        """Ensure that the cell_ids that are to be extracted are not too close to image edges

        Returns:
            list: List of cell IDs that are too close to the image border to be extracted
        """

        def _check_location(
            id, center: tuple[float, float], width: int, image_width: int, image_height: int
        ) -> int | None:
            x, y = center

            if x <= width:
                return id
            if y <= width:
                return id
            if x >= image_width - width:
                return id
            if y >= image_height - width:
                return id

            return None

        f = func_partial(
            _check_location,
            width=self.width_extraction,
            image_width=self.input_image_width,
            image_height=self.input_image_height,
        )
        results = {f(id, center) for id, center in zip(self.classes, self.px_centers, strict=True)}
        results.discard(None)

        return list(results)

    def _verbalise_extraction_info(self) -> None:
        """Print relevant information about the extraction process to the log file."""
        # print some output information
        self.log("Extraction Details:")
        self.log("--------------------------------")
        self.log(f"Number of input image channels: {self.n_image_channels}")
        self.log(f"Number of segmentation masks used during extraction: {self.n_masks}")
        self.log(f"Number of generated output images per cell: {self.n_output_channels}")
        self.log(f"Number of unique cells to extract: {self.num_classes}")
        self.log(f"Extracted Image Dimensions: {self.extracted_image_size} x {self.extracted_image_size}")
        self.log(f"Normalization of extracted images: {self.normalization}")
        self.log(f"Percentile normalization range for single-cell images: {self.normalization_range}")

    def _generate_save_index_lookup(self, class_list: list) -> None:
        """Create a lookup index indicating at which save_index each cell_id should be saved to in the HDF5 file."""
        self.save_index_lookup = pd.DataFrame(index=class_list)

    def _get_arg(self, cell_ids: list[int], centers: list[tuple[float, float]]) -> list:
        """Helper function for _generate_batched_args.

        The logic here is kept as it's own function, to allow for the simple changing fo this logic in alternative extraction workflow impelementations.
        Assumption is that provided cell_ids and centers are matched and sorted identically.

        Args:
            cell_ids: List of cell IDs to extract.
            centers: List of pixel centers for the cells to extract.

        Returns:
            list: List of arguments to be used for multiprocessing
        """
        args = list(
            zip(
                range(len(cell_ids)),
                [self.save_index_lookup.index.get_loc(x) for x in cell_ids],
                cell_ids,
                centers,
                strict=False,
            )
        )
        return args

    def _generate_batched_args(self, args: list, max_batch_size: int = 1000, min_batch_size: int = 100) -> list[list]:
        """
        Helper function to generate batched arguments for multiprocessing.
        Batched args are mini-batches of the original arguments that are used to split the processing into smaller chunks to prevent memory issues.
        The batch size for each mini batch can be configured witht he max_batch_size and min_batch_size parameters.
        The final batch size will be between the two values ensuring an equal distribution of arguments to desired number of processes.

        Args:
            args: List of arguments that need to be batched generated by self._get_arg
            max_batch_size: upper limit for the mini batch size
            min_batch_size: lower limit fr the mini batch size

        Returns:
            list: List of batched arguments.
        """

        if "max_batch_size" in self.config:
            max_batch_size = self.config["max_batch_size"]
        else:
            max_batch_size = max_batch_size

        theoretical_max = np.ceil(len(args) / self.threads)
        batch_size = np.int64(min(max_batch_size, theoretical_max))

        self.batch_size = np.int64(max(min_batch_size, batch_size))
        self.log(f"Using batch size of {self.batch_size} for multiprocessing.")

        # dynamically adjust the number of threads to ensure that we dont initiate more threads than we have arguments
        self.threads = np.int64(min(self.threads, np.ceil(len(args) / self.batch_size)))

        if self.threads != self.threads:
            self.log(f"Reducing number of threads to {self.threads} to match number of cell batches to process.")

        return [args[i : i + self.batch_size] for i in range(0, len(args), self.batch_size)]

    def _get_label_info(self, arg: tuple) -> tuple:
        """adjust the label information for the extraction based on the segmentation masks used"""
        index, save_index, cell_id, px_center = arg

        # no additional labelling required
        return (index, save_index, cell_id, px_center, None, None)

    def _save_removed_classes(self, classes: np.ndarray) -> None:
        """Save cell IDs that are too close to the image border to be extracted as a csv file.

        Args:
            classes: List of cell IDs that are too close to the image border to be extracted.
        """

        # define path where classes should be saved
        filtered_path = os.path.join(self.extraction_data_directory, self.DEFAULT_REMOVED_CLASSES_FILE)

        to_write = "\n".join([str(i) for i in list(classes)])

        with open(filtered_path, "w") as myfile:
            myfile.write(to_write)

        self.log(
            f"A total of {len(classes)} cells were too close to the image border to be extracted. Their cell_ids were saved to file {filtered_path}."
        )

    def _extract_classes(
        self, arg: tuple[int, int, int, tuple[float, float], int | None, str | None], return_results: bool = False
    ) -> None | tuple[int, np.ndarray, int]:
        """
        extract single-cell images for a given cell_id. If return_results is True, the results are returned, otherwise they are directly saved to the HDF5 file.
        If running in multi-threading mode, return_results should be set to False since concurrent access to the HDF5 file is not supported.

        Args:
            arg: a tuple containing all of the information needed to extract the specified single-cell. The tuple contains the index, the save_index, the cell_id, the pixel center, the image index, and the label_info.
            return_results (bool): Whether to return the results or save them to the HDF5 file.

        Returns:
            None: If return_results is False.
            tuple: If return_results is True. The tuple contains the save_index, the stack of images, and the cell_id.
        """
        index, save_index, cell_id, px_center, image_index, label_info = self._get_label_info(
            arg
        )  # label_info not used in base case but relevant for flexibility for other classes

        # currently this defaults to the same id for both
        # in theory scPortrait would be compatible with mapping nuclues ids to cytosol ids instead of updating the mask and then this code would need to change
        # this is a placeholder for now where this code can be implemented in the future but so that the rest of the pipeline already works with this use case
        nucleus_id = cell_id
        cytosol_id = cell_id

        ids = []
        if self.extract_nucleus_mask:
            nucleus_id = cell_id
            ids.append(nucleus_id)
        if self.extract_cytosol_mask:
            cytosol_id = cell_id
            ids.append(cytosol_id)

        # get region that should be extracted
        window_y = slice(px_center[1] - self.width_extraction, px_center[1] + self.width_extraction)
        window_x = slice(px_center[0] - self.width_extraction, px_center[0] + self.width_extraction)

        # ensure that the cell is not too close to the image edge to be extracted
        condition = [
            self.width_extraction < px_center[0],
            px_center[0] < self.input_image_width - self.width_extraction,
            self.width_extraction < px_center[1],
            px_center[1] < self.input_image_height - self.width_extraction,
        ]
        if not condition:
            raise ValueError("Cell is too close to the image edge to be extracted.")

        masks = []

        # get the segmentation masks
        for mask_ix in range(self.n_masks):
            if image_index is None:
                # nuclei_mask = self.sdata[self.nucleus_key].data[window_y, window_x].compute()
                mask = self.seg_masks[mask_ix, window_y, window_x]
            else:
                # nuclei_mask = self.sdata[self.nucleus_key].data[image_index, window_y, window_x].compute()
                mask = self.seg_masks[image_index, mask_ix, window_y, window_x]

            # modify nucleus mask to only contain the nucleus of interest and perform some morphological operations to generate a good image representation of the mask
            mask = np.where(mask == ids[mask_ix], 1, 0)
            mask = binary_fill_holes(mask)
            mask = gaussian(mask, preserve_range=True, sigma=1)

            if self.deep_debug:
                if mask.shape != (self.extracted_image_size, self.extracted_image_size):
                    print("Width of window_x", window_x.stop - window_x.start)
                    print("Width of window_y", window_y.stop - window_y.start)
                    print("Width of mask", mask.shape)
                    print("px_center", px_center)
                    print("cell_id", ids[mask_ix])
            assert mask.shape == (
                self.extracted_image_size,
                self.extracted_image_size,
            ), "Mask shape does not match extracted image size."
            masks.append(mask)

        # get the image data
        if image_index is None:
            image_data = self.image_data[:, window_y, window_x]
        else:
            image_data = self.image_data[image_index, :, window_y, window_x]

        image_data = (
            image_data * masks[-1]
        )  # always uses the last available mask, in nucleus only seg its the nucleus, if both its the cytosol, if only cytosol its also the cytosol. This always is the mask we want to use to extract the channel information

        # this needs to be performed on a per channel basis!
        images = []
        for i in range(image_data.shape[0]):
            ix = self.norm_function(image_data[i])
            images.append(ix)

        inputs = masks + images
        # masks and images have the same dtype here as both have been converted to images scaled between 0 and 1
        # ensuring the same dtype for all images is essential as otherwise they can not be saved into the same HDF5 container
        # the masks have also been slighly modified from the original binary masks to include a slight gaussian blur at the mask edges to smooth the masking of the image data
        stack = np.stack(inputs, axis=0).astype(self.DEFAULT_SINGLE_CELL_IMAGE_DTYPE)

        if self.deep_debug:
            # visualize some cells for debugging purposes
            if index % 1000 == 0:
                print(f"Cell ID: {cell_id} has center at [{px_center[0]}, {px_center[1]}]")

                fig, axs = plt.subplots(1, stack.shape[0], figsize=(2 * stack.shape[0], 2))
                for i, img in enumerate(stack):
                    axs[i].imshow(img, vmin=0, vmax=1)
                    axs[i].axis("off")
                fig.tight_layout()
                fig.show()

        if return_results:
            return save_index, stack, cell_id
        else:
            self._single_cell_data_container[save_index] = stack
            self._single_cell_index_container[save_index] = cell_id
            return None

    def _extract_classes_multi(
        self,
        arg_list: list[tuple[int, int, int, tuple[float, float], int | None, str | None]],
    ) -> list[tuple[int, np.ndarray, int]]:
        """Wrapper function to process all single-cells from a mini-batch.

        Args:
            arg_list: List of arguments to process.

        Returns:
            list: List of results from processing the mini-batch.
        """
        # set up normalization function
        self._setup_normalization()

        self.seg_masks: np.ndarray = mmap_array_from_path(self.path_seg_masks)
        self.image_data: np.ndarray = mmap_array_from_path(self.path_image_data)

        # get processing results
        results = [self._extract_classes(arg, return_results=True) for arg in arg_list]

        return results

    def _write_to_hdf5(self, results: list[tuple[int, np.ndarray, int]], hdf5_lock) -> None:
        """Function for writing results to HDF5 file in a thread-safe manner.

        Args:
            results: List of results to write to the HDF5 file.
            hdf5_lock: Lock object to ensure thread safety.
        """
        with hdf5_lock:
            with h5py.File(self.output_path, "a") as hf:
                self._single_cell_data_container: h5py.Dataset = hf[self.IMAGE_DATACONTAINER_NAME]
                self._single_cell_index_container: h5py.Dataset = hf[self.INDEX_DATACONTAINER_NAME]

                for res in results:
                    save_index, stack, cell_id = res
                    self._single_cell_data_container[save_index] = stack
                    self._single_cell_index_container[save_index] = cell_id

    def _initialize_empty_anndata(self) -> None:
        """Initialize an AnnData object to store the extracted single-cell images."""

        mask_names = np.array(self.masks, dtype="<U15")
        channel_names = np.array(self.channel_names, dtype="<U15")
        channels = np.concatenate([mask_names, channel_names])
        channel_mapping = ["mask" for x in mask_names] + ["image_channel" for x in channel_names]

        # create var object with channel names and their mapping to mask or image channels
        vars = pd.DataFrame(index=np.arange(len(channels)).astype("str"))
        vars["channels"] = channels
        vars["channel_mapping"] = channel_mapping

        # create empty obs object
        obs = pd.DataFrame(
            {self.DEFAULT_CELL_ID_NAME: np.zeros(shape=(self.num_classes), dtype=self.DEFAULT_SEGMENTATION_DTYPE)}
        )
        obs.index = obs.index.values.astype("str")

        # create anndata object
        adata = AnnData(obs=obs, var=vars)

        # add additional metadata to `uns`
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/n_cells"] = self.num_classes
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/n_channels"] = self.n_masks + self.n_image_channels
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/n_masks"] = self.n_masks
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/n_image_channels"] = self.n_image_channels
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/image_size"] = self.image_size
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/normalization"] = self.normalization
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/normalization_range_lower"] = self.normalization_range[0]
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/normalization_range_upper"] = self.normalization_range[1]
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/channel_names"] = channels
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/channel_mapping"] = np.array(channel_mapping, dtype="<U15")
        adata.uns[f"{self.DEFAULT_NAME_SINGLE_CELL_IMAGES}/compression"] = self.compression_type

        # write to file
        adata.write(self.output_path)

    def _create_output_files(self) -> None:
        """Initialize the output HDF5 results file."""

        # define shapes for the output containers
        # single cell data: [n_cells, n_masks + n_image_channels, image_size, image_size]

        single_cell_data_shape = (
            self.num_classes,
            (self.n_masks + self.n_image_channels),
            self.image_size,
            self.image_size,
        )

        self.output_path = os.path.join(self.extraction_data_directory, self.DEFAULT_EXTRACTION_FILE)
        self._initialize_empty_anndata()

        # add an empty HDF5 dataset to the obsm group of the anndata object
        with h5py.File(self.output_path, "a") as hf:
            hf.create_dataset(
                self.IMAGE_DATACONTAINER_NAME,
                shape=single_cell_data_shape,
                chunks=(1, 1, self.image_size, self.image_size),
                compression=self.compression_type,
                dtype=self.DEFAULT_SINGLE_CELL_IMAGE_DTYPE,
            )

            # add required metadata from anndata package
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["encoding-type"] = "array"
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["encoding-version"] = "0.2.0"

            # add relevant metadata to the single-cell image container
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["n_cells"] = self.num_classes
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["n_channels"] = self.n_masks + self.n_image_channels
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["n_masks"] = self.n_masks
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["n_image_channels"] = self.n_image_channels
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["image_size"] = self.image_size
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["normalization"] = self.normalization
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["normalization_range"] = self.normalization_range
            masks = [x.encode("utf-8") for x in self.masks]
            channel_names = [x.encode("utf-8") for x in self.channel_names]
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["channel_names"] = np.array(masks + channel_names)
            mapping_values = ["mask" for x in masks] + ["image_channel" for x in channel_names]
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["channel_mapping"] = np.array(
                [x.encode("utf-8") for x in mapping_values]
            )
            hf[self.IMAGE_DATACONTAINER_NAME].attrs["compression"] = self.compression_type

            self.log("Container for single-cell data created.")

    def _post_extraction_cleanup(self, vars_to_delete=None):
        """remove temporary directories and files created during extraction. Reset attributes that are no longer required."""
        # remove normalization functions becuase other subsequent multiprocessing calls will fail
        if "norm_function" in self.__dict__:
            del self.norm_function

        # delete segmentation masks and input images from self if present
        if "seg_masks" in self.__dict__:
            del self.seg_masks
        if "image_data" in self.__dict__:
            del self.image_data

        if "_single_cell_data_container" in self.__dict__:
            del self._single_cell_data_container
        if "_single_cell_index_container" in self.__dict__:
            del self._single_cell_index_container

        # remove no longer required variables
        if vars_to_delete is not None:
            self._clear_cache(vars_to_delete=vars_to_delete)

        # remove memory mapped temp files
        if os.path.exists(self.path_seg_masks):
            os.remove(self.path_seg_masks)
        if os.path.exists(self.path_image_data):
            os.remove(self.path_image_data)

        # ensure that the save_index_to_remove is deleted to clear up memory and prevent issues with subsequent calls
        if "save_index_to_remove" in self.__dict__:
            self.save_index_to_remove = []  # reinitalize as empty

        self._clear_cache()

    def _save_benchmarking_times(
        self,
        total_time: float,
        time_setup: float,
        time_arg_generation: float,
        time_data_transfer: float,
        time_extraction: float,
        rate_extraction: float,
    ) -> None:
        """Write benchmarking times to csv file.

        Args:
            total_time: Total time taken for the extraction.
            time_setup: Time taken to set up the extraction.
            time_arg_generation: Time taken to generate arguments.
            time_data_transfer: Time taken to transfer data to memory mapped arrays.
            time_extraction: Time taken to extract single cell images.
            rate_extraction: Rate of extraction.

        """
        # save benchmarking times to file
        benchmarking_path = os.path.join(self.directory, self.DEFAULT_BENCHMARKING_FILE)

        benchmarking = pd.DataFrame(
            {
                "Size of image extracted from": [
                    (
                        self.n_image_channels,
                        self.input_image_width,
                        self.input_image_height,
                    )
                ],
                "Number of classes extracted": [self.num_classes],
                "Number of masks used for extraction": [self.n_masks],
                "Size of extracted images": [self.extracted_image_size],
                "Number of threads used": [self.threads],
                "Mini_batch size": [self.batch_size],
                "Total extraction time": [total_time],
                "Time taken to set up extraction": [time_setup],
                "Time taken to generate arguments": [time_arg_generation],
                "Time taken to transfer data to memory mapped arrays": [time_data_transfer],
                "Time taken to extract single cell images": [time_extraction],
                "Rate of extraction": [rate_extraction],
            }
        )

        if os.path.exists(benchmarking_path):
            # append to existing file
            benchmarking.to_csv(benchmarking_path, mode="a", header=False, index=False)
        else:
            # create new file
            benchmarking.to_csv(benchmarking_path, index=False)
        self.log("Benchmarking times saved to file.")

    def process(
        self, partial: bool = False, n_cells: int = None, seed: int = 42, output_folder_name: str | None = None
    ) -> None:
        """
        Extracts single cell images from a segmented scPortrait project and saves the results to a standardized HDF5 file.

        Args:
            input_segmentation_path : Path of the segmentation HDF5 file. If this class is used as part of a project processing workflow, this argument will be provided automatically.
            partial: if set to True only a random subset of n_cells will be extracted.
            n_cells: Number of cells to extract if partial is set to True.
            seed: Seed for random sampling of cells for reproducibility if partial is set to True.

        Important
        ---------
        If this class is used as part of a project processing workflow, all of the arguments will be provided by the ``Project`` class based on the previous segmentation.
        The Project class will automatically provide the most recent segmentation forward together with the supplied parameters.

        Examples
        --------
        .. code-block:: python

            # After project is initialized and input data has been loaded and segmented
            project.extract()

        Notes
        -----
        The following parameters are required in the config file when running this method:

        .. code-block:: yaml

            HDF5CellExtraction:
                # threads used in multithreading
                threads: 80

                # image size in pixels
                image_size: 128

                # directory where intermediate results should be saved
                cache: "/mnt/temp/cache"
        """
        total_time_start = timeit.default_timer()

        start_setup = timeit.default_timer()

        # set up flag for partial processing
        self.partial_processing = partial
        if self.partial_processing:
            self.n_cells = n_cells
            self.seed = seed
            self.DEFAULT_LOG_NAME = "partial_processing.log"  # change log name so that the results are not written to the same log file as a complete extraction

        # run all of the extraction setup steps
        self._set_up_extraction(output_folder_name=output_folder_name)
        stop_setup = timeit.default_timer()
        time_setup = stop_setup - start_setup

        if self.partial_processing:
            self.log(f"Starting partial single-cell image extraction of {self.n_cells} cells...")
        else:
            self.log(f"Starting single-cell image extraction of {self.num_classes} cells...")

        # generate cell pairings to extract
        start_arg_generation = timeit.default_timer()

        self._generate_save_index_lookup(self.classes)
        args = self._get_arg(self.classes, self.px_centers)
        stop_arg_generation = timeit.default_timer()
        time_arg_generation = stop_arg_generation - start_arg_generation

        # convert input images to memory mapped temp arrays for faster reading
        self.log("Loading input images to memory mapped arrays...")
        start_data_transfer = timeit.default_timer()

        self.path_seg_masks = self.filehandler._load_seg_to_memmap(
            seg_name=self.masks, tmp_dir_abs_path=self._tmp_dir_path
        )
        self.path_image_data = self.filehandler._load_input_image_to_memmap(tmp_dir_abs_path=self._tmp_dir_path)

        stop_data_transfer = timeit.default_timer()
        time_data_transfer = stop_data_transfer - start_data_transfer

        if self.debug:
            self.log(
                f"Finished transferring data to memory mapped arrays. Time taken: {time_data_transfer:.2f} seconds."
            )

        # actually perform single-cell image extraction
        start_extraction = timeit.default_timer()

        if self.threads <= 1:
            # set up for single-threaded processing
            self._setup_normalization()

            self.seg_masks = mmap_array_from_path(self.path_seg_masks)
            self.image_data = mmap_array_from_path(self.path_image_data)

            with h5py.File(
                self.output_path,
                "a",
            ) as hf:
                # connect to final containers for saving computed results
                self._single_cell_data_container = hf[self.IMAGE_DATACONTAINER_NAME]
                self._single_cell_index_container = hf[self.INDEX_DATACONTAINER_NAME]

                self.log("Running in single threaded mode.")
                for arg in tqdm(args, total=len(args), desc="Extracting cell batches"):
                    self._extract_classes(arg)
        else:
            args = self._generate_batched_args(args)

            self.log(f"Running in multiprocessing mode with {self.threads} threads.")

            with mp.Manager() as manager:
                lock = manager.Lock()  # Create lock via Manager to enable sharing

                with mp.get_context("fork").Pool(
                    processes=self.threads
                ) as pool:  # both spawn and fork work but fork is faster so forcing fork here
                    for result in list(
                        tqdm(
                            pool.imap(self._extract_classes_multi, args),
                            total=len(args),
                            desc="Extracting cell batches",
                        )
                    ):
                        self._write_to_hdf5(result, lock)
                    pool.close()
                    pool.join()

        # cleanup memory and remove any no longer required variables
        del args
        stop_extraction = timeit.default_timer()

        # calculate duration
        time_extraction = stop_extraction - start_extraction
        rate = self.num_classes / time_extraction

        # generate final log entries
        self.log(f"Finished extraction in {time_extraction:.2f} seconds ({rate:.2f} cells / second)")

        if self.partial_processing:
            self.DEFAULT_LOG_NAME = "processing.log"  # change log name back to default

        self._post_extraction_cleanup()
        total_time_stop = timeit.default_timer()
        total_time = total_time_stop - total_time_start

        self._save_benchmarking_times(
            total_time=total_time,
            time_setup=time_setup,
            time_arg_generation=time_arg_generation,
            time_data_transfer=time_data_transfer,
            time_extraction=time_extraction,
            rate_extraction=rate,
        )



import torch


def precision(predictions, labels, pos_label=0):
    """
    Calculate precision for predicting class `pos_label`.

    Parameters
    ----------
    predictions : torch.Tensor
        Model predictions.
    labels : torch.Tensor
        Ground truth labels.
    pos_label : int, optional, default = 0
        The positive label for which to calculate precision.

    Returns
    -------
    precision : float
        Precision for predicting class `pos_label`.
    """

    _, max_indices = torch.max(predictions, 1)

    # bool array for all elements which equal the class label
    # all predicted positives
    all_pred_positives = max_indices == pos_label

    # masked ground truth tensor for all values predicetd to be positive
    masked = labels.masked_select(all_pred_positives)

    # true positives divided by all positives
    correct = (masked == pos_label).sum()
    precision = correct / len(masked)

    return precision


def recall(predictions, labels, pos_label=0):
    """
    Calculate recall for predicting class `pos_label`.

    Parameters
    ----------
    predictions : torch.Tensor
        Model predictions.
    labels : torch.Tensor
        Ground truth labels.
    pos_label : int, optional, default = 0
        The positive label for which to calculate precision.

    Returns
    -------
    recall : float
        Recall for predicting class `pos_label`.
    """

    _, max_indices = torch.max(predictions, 1)

    # all true positive elements
    all_labled_positives = labels == pos_label

    # masked prediction tensor for all values ground truth positive
    masked = max_indices.masked_select(all_labled_positives)

    correct = (masked == pos_label).sum()
    recall = correct / len(masked)

    return recall



import multiprocessing as mp
import os
import timeit
from functools import partial as func_partial

import matplotlib.pyplot as plt
import numpy as np
import xarray
from lmd.lib import SegmentationLoader
from scipy.sparse import coo_array
from tqdm.auto import tqdm

from scportrait.pipeline._base import ProcessingStep
from scportrait.pipeline._utils.helper import flatten


class LMDSelection(ProcessingStep):
    """
    Select single cells from a segmented sdata file and generate cutting data for the Leica LMD microscope.
    This method class relies on the functionality of the pylmd library.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._check_config()

        self.name = None
        self.cell_sets = None
        self.calibration_marker = None

        self.deep_debug = False  # flag for deep debugging by developers

    def _check_config(self):
        assert "segmentation_channel" in self.config, "segmentation_channel not defined in config"
        self.segmentation_channel_to_select = self.config["segmentation_channel"]

        # check for optional config parameters

        # this defines how large the box mask around the center of a cell is for the coordinate extraction
        # assumption is that all pixels belonging to each mask are within the box otherwise they will be cut off during cutting contour generation

        if "cell_width" in self.config:
            self.cell_radius = self.config["cell_width"]
        else:
            self.cell_radius = 100

        if "threads" in self.config:
            self.threads = self.config["threads"]
            assert self.threads > 0, "threads must be greater than 0"
            assert isinstance(self.threads, int), "threads must be an integer"
        else:
            self.threads = 10

        if "batch_size_coordinate_extraction" in self.config:
            self.batch_size = self.config["batch_size_coordinate_extraction"]
            assert self.batch_size > 0, "batch_size_coordinate_extraction must be greater than 0"
            assert isinstance(self.batch_size, int), "batch_size_coordinate_extraction must be an integer"
        else:
            self.batch_size = 100

        if "orientation_transform" in self.config:
            self.orientation_transform = self.config["orientation_transform"]
        else:
            self.orientation_transform = [[0, -1], [1, 0]]
            self.config["orientation_transform"] = (
                self.orientation_transform
            )  # ensure its also in config so its passed on to the segmentation loader

        if "processes_cell_sets" in self.config:
            self.processes_cell_sets = self.config["processes_cell_sets"]
            assert self.processes_cell_sets > 0, "processes_cell_sets must be greater than 0"
            assert isinstance(self.processes_cell_sets, int), "processes_cell_sets must be an integer"
        else:
            self.processes_cell_sets = 1

    def _setup_selection(self):
        # configure name of extraction
        if self.name is None:
            try:
                name = "_".join([cell_set["name"] for cell_set in self.cell_sets])
            except KeyError:
                Warning("No name provided for the selection. Will use default name.")
                name = "selected_cells"
        else:
            name = self.name

        # create savepath
        savename = name.replace(" ", "_") + ".xml"
        self.savepath = os.path.join(self.directory, savename)

        # check that the segmentation label exists
        assert (
            self.segmentation_channel_to_select in self.project.filehandler.get_sdata()._shared_keys
        ), f"Segmentation channel {self.segmentation_channel_to_select} not found in sdata."

    def __get_coords(
        self, cell_ids: list, centers: list[tuple[int, int]], width: int = 60
    ) -> list[tuple[int, np.ndarray]]:
        results = []

        _sdata = self.project.filehandler.get_sdata()

        # get correct scale of segmentaiton mask
        seg_mask = _sdata[self.segmentation_channel_to_select]
        if isinstance(seg_mask, xarray.DataArray):
            seg_mask = seg_mask.data
        elif isinstance(seg_mask, xarray.DataTree):
            seg_mask = seg_mask.scale0.image

        for i, _id in enumerate(cell_ids):
            values = centers[i]

            x_start = np.max([int(values[0]) - width, 0])
            y_start = np.max([int(values[1]) - width, 0])

            x_end = x_start + width * 2
            y_end = y_start + width * 2

            _cropped = seg_mask[slice(x_start, x_end), slice(y_start, y_end)].compute()

            # optional plotting output for deep debugging
            if self.deep_debug:
                if self.threads == 1:
                    plt.figure()
                    plt.imshow(_cropped)
                    plt.show()
                else:
                    raise ValueError("Deep debug is not supported with multiple threads.")

            sparse = coo_array(_cropped == _id)

            if (
                0 in sparse.coords[0]
                or 0 in sparse.coords[1]
                or width * 2 - 1 in sparse.coords[0]
                or width * 2 - 1 in sparse.coords[1]
            ):
                Warning(
                    f"Cell {i} with id {_id} is potentially not fully contained in the bounding mask. Consider increasing the value for the 'cell_width' parameter in your config."
                )

            x = sparse.coords[0] + x_start
            y = sparse.coords[1] + y_start

            results.append((_id, np.array(list(zip(x, y, strict=True)))))

        return results

    def _get_coords_multi(self, width: int, arg: tuple[list[int], np.ndarray]) -> list[tuple[int, np.ndarray]]:
        cell_ids, centers = arg
        results = self.__get_coords(cell_ids, centers, width)
        return results

    def _get_coords(
        self, cell_ids: list, centers: list[tuple[int, int]], width: int = 60, batch_size: int = 100, threads: int = 10
    ) -> dict[int, np.ndarray]:
        # create batches
        n_batches = int(np.ceil(len(cell_ids) / batch_size))
        slices = [(i * batch_size, i * batch_size + batch_size) for i in range(n_batches - 1)]
        slices.append(((n_batches - 1) * batch_size, len(cell_ids)))

        batched_args = [(cell_ids[start:end], centers[start:end]) for start, end in slices]

        f = func_partial(self._get_coords_multi, width)

        if (
            threads == 1
        ):  # if only one thread is used, the function is called directly to avoid the overhead of multiprocessing
            results = [f(arg) for arg in batched_args]
        else:
            with mp.get_context(self.context).Pool(processes=threads) as pool:
                results = list(
                    tqdm(
                        pool.imap(f, batched_args),
                        total=len(batched_args),
                        desc="Processing cell batches",
                    )
                )
                pool.close()
                pool.join()

        results = flatten(results)  # type: ignore
        return dict(results)  # type: ignore

    def _get_cell_ids(self, cell_sets: list[dict]) -> list[int]:
        cell_ids = []
        for cell_set in cell_sets:
            if "classes" in cell_set:
                cell_ids.extend(cell_set["classes"])
            else:
                Warning(f"Cell set {cell_set['name']} does not contain any classes.")
        return cell_ids

    def _get_centers(self, cell_ids: list[int]) -> list[tuple[int, int]]:
        _sdata = self.project.filehandler.get_sdata()
        centers = _sdata[f"{self.DEFAULT_CENTERS_NAME}_{self.segmentation_channel_to_select}"].compute()
        centers = centers.loc[cell_ids, :]
        return centers[
            ["y", "x"]
        ].values.tolist()  # needs to be returned as yx to match the coordinate system as saved in spatialdataobjects

    def _post_processing_cleanup(self, vars_to_delete: list | None = None):
        if vars_to_delete is not None:
            self._clear_cache(vars_to_delete=vars_to_delete)

        self._clear_cache()

    def process(
        self,
        cell_sets: list[dict],
        calibration_marker: np.array,
        name: str | None = None,
    ):
        """
        Process function for selecting cells and generating their XML.
        Under the hood this method relies on the pylmd library and utilizies its `SegmentationLoader` Class.

        Args:
            cell_sets (list of dict): List of dictionaries containing the sets of cells which should be sorted into a single well. Mandatory keys for each dictionary are: name, classes. Optional keys are: well.
            calibration_marker (numpy.array): Array of size ‘(3,2)’ containing the calibration marker coordinates in the ‘(row, column)’ format.
            name (str, optional): Name of the output file. If not provided, the name will be generated based on the names of the cell sets or if also not specified set to "selected_cells".

        Example:

            .. code-block:: python

                # Calibration marker should be defined as (row, column).
                marker_0 = np.array([-10, -10])
                marker_1 = np.array([-10, 1100])
                marker_2 = np.array([1100, 505])

                # A numpy Array of shape (3, 2) should be passed.
                calibration_marker = np.array([marker_0, marker_1, marker_2])

                # Sets of cells can be defined by providing a name and a list of classes in a dictionary.
                cells_to_select = [{"name": "dataset1", "classes": [1, 2, 3]}]

                # Alternatively, a path to a csv file can be provided.
                # If a relative path is provided, it is accessed relativ to the projects base directory.
                cells_to_select += [{"name": "dataset2", "classes": "segmentation/class_subset.csv"}]

                # If desired, wells can be passed with the individual sets.
                cells_to_select += [{"name": "dataset3", "classes": [4, 5, 6], "well": "A1"}]

                project.select(cells_to_select, calibration_marker)

        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                LMDSelection:
                    #the number of threads with which multithreaded tasks should be executed
                    threads: 10

                    # the number of parallel processes to use for generation of cell sets each set
                    # will processed with the designated number of threads
                    processes_cell_sets: 1

                    # defines the channel used for generating cutting masks
                    # segmentation.hdf5 => labels => segmentation_channel
                    # When using WGA segmentation:
                    #    0 corresponds to nuclear masks
                    #    1 corresponds to cytosolic masks.
                    segmentation_channel: 0

                    # dilation of the cutting mask in pixel
                    shape_dilation: 10

                    # Cutting masks are transformed by binary dilation and erosion
                    binary_smoothing: 3

                    # number of datapoints which are averaged for smoothing
                    # the number of datapoints over an distance of n pixel is 2*n
                    convolution_smoothing: 25

                    # fold reduction of datapoints for compression
                    rdp: 0.7

                    # Optimization of the cutting path inbetween shapes
                    # optimized paths improve the cutting time and the microscopes focus
                    # valid options are ["none", "hilbert", "greedy"]
                    path_optimization: "hilbert"

                    # Paramter required for hilbert curve based path optimization.
                    # Defines the order of the hilbert curve used, which needs to be tuned with the total cutting area.
                    # For areas of 1 x 1 mm we recommend at least p = 4,  for whole slides we recommend p = 7.
                    hilbert_p: 7

                    # Parameter required for greedy path optimization.
                    # Instead of a global distance matrix, the k nearest neighbours are approximated.
                    # The optimization problem is then greedily solved for the known set of nearest neighbours until the first set of neighbours is exhausted.
                    # Established edges are then removed and the nearest neighbour approximation is recursivly repeated.
                    greedy_k: 20

                    # The LMD reads coordinates as integers which leads to rounding of decimal places.
                    # Points spread between two whole coordinates are therefore collapsed to whole coordinates.
                    # This can be mitigated by scaling the entire coordinate system by a defined factor.
                    # For a resolution of 0.6 um / px a factor of 100 is recommended.
                    xml_decimal_transform: 100

                    # Overlapping shapes are merged based on a nearest neighbour heuristic.
                    # All selected shapes closer than distance_heuristic pixel are checked for overlap.
                    distance_heuristic: 300

        """

        self.log("Selection process started.")

        self.name = name
        self.cell_sets = cell_sets
        self.calibration_marker = calibration_marker

        self._setup_selection()

        start_time = timeit.default_timer()
        cell_ids = self._get_cell_ids(cell_sets)
        centers = self._get_centers(cell_ids)

        coord_index = self._get_coords(
            cell_ids=cell_ids, centers=centers, width=self.cell_radius, batch_size=self.batch_size, threads=self.threads
        )
        self.log(f"Coordinate lookup index calculation took {timeit.default_timer() - start_time} seconds.")

        sl = SegmentationLoader(
            config=self.config,
            verbose=self.debug,
            processes=self.config["processes_cell_sets"],
        )

        shape_collection = sl(None, self.cell_sets, self.calibration_marker, coords_lookup=coord_index)

        if self.debug:
            shape_collection.plot(calibration=True)
            shape_collection.stats()

        shape_collection.save(self.savepath)

        self.log(f"Saved output at {self.savepath}")

        # perform post processing cleanup
        self._post_processing_cleanup(vars_to_delete=[shape_collection, sl, coord_index])



"""
parse
====================================

Contains functions to parse imaging data aquired on an OperaPhenix or Operetta into usable formats for downstream pipelines.
"""

import os
import platform
import re
import shutil
import sys
import warnings
import xml.etree.ElementTree as ET
from datetime import datetime
from pathlib import PosixPath

import numpy as np
import pandas as pd
from tifffile import imread, imwrite
from tqdm.auto import tqdm


def _get_child_name(elem):
    return elem.split("}")[1]


class PhenixParser:
    """
    A class to parse and manage image data from Phenix experiments.
    """

    def __init__(
        self,
        experiment_dir: str | PosixPath,
        flatfield_exported: bool = True,
        use_symlinks: bool = True,
        compress_rows: bool = False,
        compress_cols: bool = False,
        overwrite: bool = False,
    ) -> None:
        """
        Args:
            experiment_dir: The directory containing the exported phenxi experiment
            flatfield_exported: Whether flatfield corrected images were exported.
            use_symlinks: Whether to use symbolic links for parsed images.
            compress_rows: Whether to compress rows in the parsed images.
            compress_cols: Whether to compress columns in the parsed images.
            overwrite: Whether to overwrite existing files during the parsing process.
        """
        self.experiment_dir = experiment_dir
        self.export_symlinks = use_symlinks
        self.flatfield_status = flatfield_exported
        self.compress_rows = compress_rows
        self.compress_cols = compress_cols
        self.overwrite = overwrite

        if self.compress_rows:
            print(
                "The rows found in the phenix layout will be compressed into one row after parsing the images, r and c indicators will be adjusted accordingly."
            )
        if self.compress_cols:
            print(
                "The wells found in the phenix layout will be compressed into one column after parsing the images, r and c indicators will be adjusted accordingly."
            )

        self.xml_path = self._get_xml_path()
        self.image_dir = self._get_input_dir()
        self.channel_lookup = self._get_channel_metadata(self.xml_path)
        self.metadata: None | pd.DataFrame = None
        self.missing_images_copy: list[str] = []
        self.outdirs: dict[str, str] = {}

    def _get_xml_path(self) -> str | PosixPath:
        """Automatically gets the path to the XML file containing metadata."""

        # directory depends on if flatfield images were exported or not
        # these generated folder structures are hard coded during phenix export, do not change

        if self.flatfield_status:
            index_file_names = ["Index.xml", "Index.ref.xml"]
            for index_file_name in index_file_names:
                index_file = os.path.join(self.experiment_dir, "Images", index_file_name)
                if os.path.isfile(index_file):
                    break
        else:
            index_file_names = ["Index.xml", "Index.idx.xml"]
            for index_file_name in index_file_names:
                index_file = os.path.join(self.experiment_dir, "Images", index_file_name)
                if os.path.isfile(index_file):
                    break

        # perform sanity check if file exists else exit
        if not os.path.isfile(index_file):
            sys.exit(f"Can not find index file at path: {index_file}")

        return index_file

    def _get_input_dir(self) -> str | PosixPath:
        """Automatically get the subfolder where the exported image files are located."""
        # directory depends on if flatfield images were exported or not
        # these generated folder structures are hard coded during phenix export, do not change
        if self.flatfield_status:
            input_dir = os.path.join(self.experiment_dir, "Images", "flex")
        else:
            input_dir = os.path.join(self.experiment_dir, "Images")

        # perform sanity check if file exists else exit
        if not os.path.isdir(input_dir):
            sys.exit(f"Can not find directory containing images to parse: {input_dir}")

        return input_dir

    def _define_outdir(self, name: str = "parsed_images") -> None:
        """Create output directory for parsed images.

        Args:
            name: Name of the output directory.
        """

        self.outdirs[name] = f"{self.experiment_dir}/{name}"

        # if output directory did not exist create it
        if not os.path.isdir(self.outdirs[name]):
            os.makedirs(self.outdirs[name])

    def _get_channel_metadata(self, xml_path: str | PosixPath) -> pd.DataFrame:
        """Parse channel metadata from Index.ref.xml.

        Args:
            xml_path: Path to the XML file containing metadata.

        Returns:
            a lookup table for channel names and IDs
        """
        index_file = xml_path

        # Read and parse the XML file
        with open(index_file) as f:
            content = f.read()

        # Use regex to detect pattern
        pattern = r"<ChannelName>.*?</ChannelName>|<ChannelID>.*?</ChannelID>"
        results = re.findall(pattern, content, re.MULTILINE)
        results = [x.strip() for x in results]

        # Extract channel IDs and names
        channel_ids = [x.split(">")[1].split("<")[0] for x in results if x.startswith("<ChannelID")]
        channel_names = [
            x.split(">")[1].split("<")[0].replace(" ", "") for x in results if x.startswith("<ChannelName")
        ]

        # Process
        channel_ids = list(set(channel_ids))
        channel_ids.sort()
        channel_names = channel_names[0 : len(channel_ids)]

        # Create DataFrame
        lookup = pd.DataFrame({"id": list(channel_ids), "label": list(channel_names)})

        print("Experiment contains the following image channels: ")
        print(lookup, "\n")

        # Save lookup file to csv
        lookup.to_csv(f"{self.experiment_dir}/channel_lookuptable.csv")
        print(f"Channel Lookup table saved to file at {self.experiment_dir}/channel_lookuptable.csv\n")

        # Set channel names
        self.channel_names = channel_names

        return lookup

    def _read_phenix_xml(self, xml_path: str | PosixPath) -> pd.DataFrame:
        """Read and parse the XML file containing metadata from a Phenix experiment.
        Returns the relevant metadata in a DataFrame.

        Args:
            xml_path: Path to the XML file containing metadata.

        Returns:
            a Dataframe containing the metadata of the experiment.
        """
        # initialize lists to save results into
        rows = []
        cols = []
        fields = []
        planes = []
        channel_ids = []
        channel_names = []
        flim_ids = []
        timepoints = []
        x_positions = []
        y_positions = []
        times = []
        url = []

        # extract information from XML file
        tree = ET.parse(xml_path)
        root = tree.getroot()

        # get Harmony Version number
        namespace = root.tag.split("}")[0].strip("{")
        if "HarmonyV5" in namespace:
            version = "HarmonyV5"
            self.harmony_version = version
        elif "HarmonyV7" in namespace:
            version = "HarmonyV7"
            self.harmony_version = version
        else:
            raise ValueError(
                f"Found a currently unsupported version number {namespace}. Please contact the developers with this example."
            )

        print(f"Parsing XML file from {version} version.")

        for i, child in enumerate(root):
            if _get_child_name(child.tag) == "Images":
                images = root[i]

        for image in images:
            for _ix, child in enumerate(image):
                tag = _get_child_name(child.tag)
                if tag == "Row":
                    rows.append(child.text)
                    continue
                elif tag == "Col":
                    cols.append(child.text)
                    continue
                elif tag == "FieldID":
                    fields.append(child.text)
                    continue
                elif tag == "PlaneID":
                    planes.append(child.text)
                    continue
                elif tag == "ChannelID":
                    channel_ids.append(child.text)
                    continue
                elif tag == "FlimID":
                    flim_ids.append(child.text)
                    continue
                elif tag == "TimepointID":
                    timepoints.append(child.text)
                    continue
                elif tag == "PositionX":
                    x_positions.append(child.text)
                    continue
                elif tag == "PositionY":
                    y_positions.append(child.text)
                    continue
                elif tag == "AbsTime":
                    times.append(child.text)
                    continue
                elif tag == "URL":
                    url.append(child.text)
                    continue
                else:
                    continue

        rows = [str(x).zfill(2) for x in rows]
        cols = [str(x).zfill(2) for x in cols]
        fields = [str(x).zfill(2) for x in fields]
        planes = [str(x).zfill(2) for x in planes]
        timepoints = [int(x) + 1 for x in timepoints]

        # get channelnames
        lookup_dict = self.channel_lookup.set_index("id").to_dict()["label"]
        channel_names = [lookup_dict[channel_id] for channel_id in channel_ids]
        channel_names = [
            x.replace(" ", "") for x in channel_names
        ]  # ensure no extra spaces are contained in the generated files

        # get file names of single-tif files on disk
        # the structure of the file names is dependent on the Harmony Version used to export them
        image_names = []
        if version == "HarmonyV5":
            for row, col, field, plane, channel_id, timepoint, flim_id in zip(
                rows, cols, fields, planes, channel_ids, timepoints, flim_ids, strict=False
            ):
                image_names.append(f"r{row}c{col}f{field}p{plane}-ch{channel_id}sk{timepoint}fk1fl{flim_id}.tiff")
        elif version == "HarmonyV7":
            timepoints = [(x - 1) for x in timepoints]
            if self.flatfield_status:
                for (
                    row,
                    col,
                    field,
                    plane,
                    channel_id,
                    timepoint,
                ) in zip(rows, cols, fields, planes, channel_ids, timepoints, strict=False):
                    image_names.append(f"r{row}c{col}f{field}p{plane}-ch{channel_id}t{str(timepoint).zfill(2)}.tiff")
            else:
                for row, col, field, plane, channel_id, timepoint, flim_id in zip(
                    rows, cols, fields, planes, channel_ids, timepoints, flim_ids, strict=False
                ):
                    image_names.append(f"r{row}c{col}f{field}p{plane}-ch{channel_id}sk{timepoint}fk1fl{flim_id}.tiff")

        # convert date/time into useful format
        dates = [x.split("T")[0] for x in times]
        _times = [x.split("T")[1] for x in times]
        _times = [(x.split("+")[0].split(".")[0] + "+" + x.split("+")[1].replace(":", "")) for x in _times]
        time_final: list[str] = [x + " " + y for x, y in zip(dates, _times, strict=False)]

        datetime_format = "%Y-%m-%d %H:%M:%S%z"
        time_unix: list[float] = [datetime.timestamp(datetime.strptime(x, datetime_format)) for x in time_final]

        # update file name if flatfield exported images are to be used
        # syntax here is somewhat complicated because export from harmony is not consistent even within the same version,
        # sometimes flex_ is prepended sometimes not (this has to do with which level of flatfield correction is available due to the number of tiles images)
        if self.flatfield_status:
            _image_names = [f"flex_{x}" for x in image_names]
            status = False
            for img in _image_names:
                path = os.path.join(self.image_dir, img)
                if os.path.exists(path):
                    status = True
                    break
            if status:
                image_names = _image_names

        df = pd.DataFrame(
            {
                "filename": image_names,
                "Row": rows,
                "Well": cols,
                "Zstack": planes,
                "Timepoint": timepoints,
                "X": x_positions,
                "Y": y_positions,
                "date": dates,
                "time": _times,
                "unix_time": time_unix,
                "Channel": channel_names,
            }
        )

        # define path where to find raw image files
        df["source"] = self.image_dir
        df["filename"] = df["filename"].astype(str)  # ensure this is a string otherwise it can cause issues later on
        df["date"] = df["date"].astype(str)
        df["time"] = df["time"].astype(str)
        df["Channel"] = df["Channel"].astype(str)

        return df

    def _get_phenix_metadata(self) -> pd.DataFrame:
        """Helper function to get metadata from Phenix XML file."""
        return self._read_phenix_xml(self.xml_path)

    def _generate_new_filenames(self, metadata: pd.DataFrame) -> pd.DataFrame:
        """Generate the new files names for each image post parsing.

        Args:
            metadata: DataFrame containing the metadata of the experiment. Generated by _get_phenix_metadata.

        Returns:
            an updated version of the input metadata which contains the new file names.
        """
        # convert position values to numeric to ensure proper sorting
        metadata["X"] = [float(x) for x in metadata.X]
        metadata["Y"] = [float(x) for x in metadata.Y]

        # convert X_positions into row and col values
        metadata["X_pos"] = None
        X_values = metadata.X.value_counts().index.to_list()
        X_values = np.sort(X_values)
        for i, x in enumerate(X_values):
            metadata.loc[metadata.X == x, "X_pos"] = i

        # get y positions
        metadata["Y_pos"] = None
        Y_values = metadata.Y.value_counts().index.to_list()
        Y_values.sort()  # ensure that the values are numeric and not string

        for i, y in enumerate(Y_values):
            metadata.loc[metadata.Y == y, "Y_pos"] = i

        # get number of rows and wells and adjust labelling if specific entries need to be compressed
        wells = metadata.Well.value_counts().index.to_list()
        rows = metadata.Row.value_counts().index.to_list()

        wells.sort()
        rows.sort(reverse=True)  # invert because we need to start assembling from bottom left

        if self.compress_rows:
            metadata["orig_Row"] = metadata["Row"]
            for well in wells:
                select_row_name = rows[0]
                for i, row in enumerate(rows):
                    if i == 0:
                        continue
                    else:
                        # get current highest index
                        max_y = metadata.loc[((metadata.Well == well) & (metadata.Row == select_row_name))].Y_pos.max()
                        # add current highest index to existing index
                        metadata.loc[(metadata.Well == well) & (metadata.Row == row), "Y_pos"] = (
                            metadata.loc[(metadata.Well == well) & (metadata.Row == row), "Y_pos"] + int(max_y) + 1
                        )
                        # update row name
                        metadata.loc[(metadata.Well == well) & (metadata.Row == row), "Row"] = select_row_name

                metadata.loc[:, "Row"] = rows[-1]  # update nomenclature to start with the row 01

        if self.compress_cols:
            metadata["orig_Well"] = metadata["Well"]
            for i, well in enumerate(wells):
                if i == 0:
                    continue
                else:
                    max_x = metadata.loc[(metadata.Well == wells[0])].X_pos.max()
                    metadata.loc[(metadata.Well == well), "X_pos"] = (
                        metadata.loc[(metadata.Well == well), "X_pos"] + int(max_x) + 1
                    )
                    metadata.loc[(metadata.Well == well), "Well"] = wells[0]

        metadata.X_pos = [str(int(x)).zfill(3) for x in metadata.X_pos]
        metadata.Y_pos = [str(int(x)).zfill(3) for x in metadata.Y_pos]
        metadata.Timepoint = [str(x).zfill(3) for x in metadata.Timepoint]
        metadata.Zstack = [str(x).zfill(2) for x in metadata.Zstack]

        # generate new file names
        for i in range(metadata.shape[0]):
            _row = metadata.loc[i, :]
            name = f"Timepoint{_row.Timepoint}_Row{_row.Row}_Well{_row.Well}_{_row.Channel}_zstack{_row.Zstack}_r{_row.Y_pos}_c{_row.X_pos}.tif"
            name = name
            metadata.loc[i, "new_file_name"] = name

        return metadata

    def _get_tile_id(self, string: str) -> str:
        """Helper function to extract tile id from filename.

        Args:
            string: Filename of the image.

        Returns:
            The tile id extracted from the filename in the format r<row>_c<col>.
        """
        pattern = r"_r(\d+)_c(\d+)\.tif"
        match = re.search(pattern, string)
        if match:
            row = match.group(1)
            col = match.group(2)
            return f"r{row}_c{col}"
        else:
            return None

    def generate_metadata(self) -> pd.DataFrame:
        """Helper function to generate metadata for the phenix experiment. Includes generating new file names for each image."""
        metadata = self._get_phenix_metadata()
        metadata_new = self._generate_new_filenames(metadata)

        # save results to self for easy access
        self.metadata = metadata_new

        return metadata_new

    def check_for_missing_files(self, metadata: pd.DataFrame = None, return_values: bool = False) -> list | None:
        """Check for missing images in the experiment.
        Stitching always requires a full rectangular grid of images, so any missing images (e.g. due to focus failures need to be identifed and replaced with black images).

        Args:
            metadata: DataFrame containing the metadata of the experiment. Generated by _get_phenix_metadata.
            return_values: Whether to return the list of missing images or only save them in self.missing_images

        Returns:
            List of missing images in the experiment if return_values is True. Otherwise saves the missing images in self.missing_images.

        """

        def _generate_missing_file_names(x_positions, y_positions, timepoint, row, well, channels, zstacks):
            """Helper function to generate missing file names given x_positions and y_positions."""

            _missing_tiles: list[str] = []

            for channel in channels:
                for zstack in zstacks:
                    for x_pos in x_positions:
                        for y_pos in y_positions:
                            _missing_tiles.append(
                                f"Timepoint{timepoint}_Row{row}_Well{well}_{channel}_zstack{zstack}_r{y_pos}_c{x_pos}.tif"
                            )
            return _missing_tiles

        # check if metadata has been passed or is already calculated, else repeat calculation
        if metadata is None:
            if "metadata" in self.__dict__:
                metadata = self.metadata
                if "new_file_name" not in metadata.columns:
                    metadata = self._generate_new_filenames(metadata)
            else:
                metadata = self.generate_metadata()
                metadata = self._generate_new_filenames(metadata)

        # get unique values for each category describing the imaging experiment
        channels = np.unique(metadata.Channel)
        zstacks = np.unique(metadata.Zstack)
        rows = np.unique(metadata.Row)
        wells = np.unique(metadata.Well)
        timepoints = np.unique(metadata.Timepoint)

        # all X and Y pos values need to be there
        y_range = np.unique(metadata.Y_pos)
        x_range = np.unique(metadata.X_pos)

        # check to ensure that no y_range or x_range value is missing
        _y_range = [int(x) for x in y_range]
        y_range = [str(x).zfill(3) for x in np.arange(min(_y_range), max(_y_range) + 1)]
        _x_range = [int(x) for x in x_range]
        x_range = [str(x).zfill(3) for x in np.arange(min(_x_range), max(_x_range) + 1)]

        # this will not catch missing tiles were an entire row or column is missing
        missing_tiles: list[str] = []
        print("Checking for missing images...")

        for timepoint in timepoints:
            _df = metadata[metadata.Timepoint == timepoint]
            for row in rows:
                __df = _df[_df.Row == row]
                if __df.empty:
                    Warning(f"Entire row {row} is missing for timepoint {timepoint}.")
                    for well in wells:
                        missing_tiles = missing_tiles + _generate_missing_file_names(
                            x_range, y_range, timepoint, row, well, channels, zstacks
                        )

                for well in wells:
                    ___df = __df[__df.Well == well]

                    if ___df.empty:
                        Warning(f"Entire well {well} is missing for timepoint {timepoint}.")
                        missing_tiles = missing_tiles + _generate_missing_file_names(
                            x_range, y_range, timepoint, row, well, channels, zstacks
                        )
                        continue

                    for x_pos in x_range:
                        _check = ___df[___df.X_pos == x_pos]
                        _y_pos = [y_pos for y_pos in y_range if y_pos not in set(_check.Y_pos)]

                        if len(_y_pos) > 0:
                            missing_tiles = missing_tiles + _generate_missing_file_names(
                                [x_pos], _y_pos, timepoint, row, well, channels, zstacks
                            )

                    for y_pos in y_range:
                        _check = ___df[___df.Y_pos == y_pos]
                        _x_pos = [x_pos for x_pos in x_range if x_pos not in set(_check.X_pos)]

                        if len(_y_pos) > 0:
                            missing_tiles = missing_tiles + _generate_missing_file_names(
                                _x_pos, [y_pos], timepoint, row, well, channels, zstacks
                            )

        if self.harmony_version == "HarmonyV7":
            missing_tiles.extend(self.missing_images_copy)  # add additional missing images from copy process
            self.missing_images = missing_tiles

        elif self.harmony_version == "HarmonyV5":
            self.missing_images = missing_tiles

        if len(self.missing_images) == 0:
            print("No missing tiles found.")
        else:
            # get size of missing images that need to be replaced
            image = None
            for source, file in zip(metadata["source"], metadata["filename"], strict=True):
                path = os.path.join(source, file)
                if os.path.exists(path):
                    image = imread(path)
                    break
            if image is None:
                raise ValueError(
                    f"Could not read any input images to determine size. Tried finding images at the following path: {path}"
                )
            image[:] = 0
            self.black_image = image

            print(f"The found missing tiles that need to be replaced with black images of the size {image.shape}.")

        if return_values:
            return missing_tiles
        else:
            return None

    def replace_missing_images(self) -> None:
        """Replace missing images with black images of the same size."""
        # calculate missing images if not already done
        if "missing_images" not in self.__dict__:
            self.check_for_missing_files()

        # initialize output directory if not already done
        if self.outdirs["parsed_images"] not in self.__dict__:
            self._define_outdir(name="parsed_images")

        # if there are missing images replace them with black images
        if len(self.missing_images) > 0:
            for missing_image in self.missing_images:
                print(f"Creating black image with name: {missing_image}")
                imwrite(os.path.join(self.outdirs["parsed_images"], missing_image), self.black_image)

            print(
                f"All missing images successfully replaced with black images of the dimension {self.black_image.shape}"
            )

    def _define_copy_functions(self) -> None:
        """Define function for copying depending on if symlinks should be used or not"""
        if self.export_symlinks:

            def copyfunction(input, output):
                try:
                    if platform.system() == "Windows":
                        warnings.warn(
                            "\n\nWindows detected as platform. Symlinks cannot be used on Windows. Using hard links instead.\n\n",
                            stacklevel=2,
                        )
                        # On Windows, use hard links when symlinks are requested
                        if not os.path.exists(output):
                            os.link(input, output)
                    else:
                        # Unix symlink behavior
                        os.symlink(input, output)
                except OSError as e:
                    print("Error: ", e)
                    return ()
        else:

            def copyfunction(input, output):
                try:
                    # Create destination directory if it doesn't exist
                    os.makedirs(os.path.dirname(output), exist_ok=True)
                    shutil.copyfile(input, output)
                except OSError as e:
                    print("Error: ", e)
                    return ()

        self.copyfunction = copyfunction

    def _copy_files(self, metadata: pd.DataFrame) -> None:
        """
        Copy files from the source directory to the output directory. The new file names are defined in the metadata.

        Args:
            metadata : parsed metadata with at least the following expected columns: filename, new_file_name, source, dest
        """
        print("Starting copy process...")
        self._define_copy_functions()
        # actually perform the copy process
        for old, new, source, dest in tqdm(
            zip(
                metadata.filename.tolist(),
                metadata.new_file_name.tolist(),
                metadata.source.tolist(),
                metadata.dest.tolist(),
                strict=False,
            ),
            total=len(metadata.new_file_name.tolist()),
            desc="Copying files",
        ):
            # define old and new paths for copy process
            old_path = os.path.join(source, old)
            new_path = os.path.join(dest, new)
            # check if old path exists
            if os.path.exists(old_path):
                if os.path.exists(new_path):
                    if self.overwrite:
                        os.remove(new_path)
                        self.copyfunction(old_path, new_path)
                    else:
                        self.copyfunction(old_path, new_path)
                else:
                    self.copyfunction(old_path, new_path)

            else:
                if self.harmony_version == "HarmonyV5":
                    print("Error: ", old_path, "not found.")
                elif self.harmony_version == "HarmonyV7":
                    self.missing_images_copy.append(new)
        print("Copy process completed.")

    def _save_metadata(self, metadata: pd.DataFrame) -> None:
        """Save metadata used to parse images to a csv file."""
        metadata.to_csv(f"{self.experiment_dir}/metadata_image_parsing.csv")
        print(f"Metadata used to parse images saved to file {self.experiment_dir}/metadata_image_parsing.csv")

    def parse(self) -> None:
        """Complete parsing of phenix experiment including checking for and replacing missing images."""
        # create output directory
        self._define_outdir(name="parsed_images")

        # get metadata for the images we want to parse
        metadata = self.generate_metadata()

        # set destination for copying
        metadata["dest"] = self.outdirs["parsed_images"]

        # copy/link the images to their new names
        self._copy_files(metadata=metadata)

        # check for missing images and replace them
        self.check_for_missing_files(metadata=metadata)
        self.replace_missing_images()
        self._save_metadata(metadata)

    def sort_wells(self, sort_tiles: bool = False) -> None:
        """Sorts parsed images according to their well.

        Generates a folder tree where each well has its own folder containing all images from that well.
        If sort_tiles = True an additional layer will be added to the tree where all images obtained from the same FOV are sorted into a unique subfolder.

        Args:
            sort_tiles: if the images should be sorted into individual directories according to FOV in addition to well, by default False
        """

        # create output directory
        self._define_outdir(name="sorted_wells")

        # get all new file names
        if "metdata" in self.__dict__:
            metadata = self.metadata
        else:
            metadata = self.generate_metadata()

        metadata["tiles"] = [self._get_tile_id(x) for x in metadata.new_file_name.to_list()]

        # get unique rows, wells and tiles
        timepoints = list(set(metadata.Timepoint.to_list()))
        rows = list(set(metadata.Row.to_list()))
        wells = list(set(metadata.Well.to_list()))
        tiles = list(set(metadata.tiles.to_list()))

        print("Found the following image specs: ")
        print("\t Timepoints: ", timepoints)
        print("\t Rows: ", rows)
        print("\t Wells: ", wells)
        if sort_tiles:
            print("\t Tiles: ", tiles)  # only print if these folders should be created
            # update metadata to include destination for each tile
            metadata["dest"] = [
                os.path.join(self.outdirs["sorted_wells"], f"row{row}_well{well}", tile)
                for row, well, tile in zip(metadata.Row, metadata.Well, metadata.tiles, strict=False)
            ]
        else:
            metadata["dest"] = [
                os.path.join(self.outdirs["sorted_wells"], f"row{row}_well{well}")
                for row, well in zip(metadata.Row, metadata.Well, strict=False)
            ]

        # unique directories for each tile
        unique_dirs = list(set(metadata.dest.to_list()))

        for _dir in unique_dirs:
            if not os.path.exists(_dir):
                os.makedirs(_dir)

        # copy/link the images to their new names
        self._copy_files(metadata=metadata)

    def sort_timepoints(self, sort_wells: bool = False) -> None:
        """Sorts parsed images according to their timepoint.

        Generates a folder tree where each timepoint has its own folder containing all images captured at that timepoint.
        If sort_wells = True an additional layer will be added to the tree where all images obtained from the same well are sorted into a unique subfolder according to timepoint.

        Args:
            sort_wells: if the images should be sorted into individual directories according to well in addition to timepoint, by default False
        """

        # create output directory
        self._define_outdir(name="sorted_timepoints")

        # get all new file names
        if "metdata" in self.__dict__:
            metadata = self.metadata
        else:
            metadata = self.generate_metadata()

        metadata["tiles"] = [self._get_tile_id(x) for x in metadata.new_file_name.to_list()]

        # get unique rows, wells and tiles
        rows = list(set(metadata.Row.to_list()))
        wells = list(set(metadata.Well.to_list()))
        tiles = list(set(metadata.tiles.to_list()))
        timepoints = list(set(metadata.Timepoint.to_list()))

        print("Found the following image specs: ")
        print("\t Timepoints: ", timepoints)
        print("\t Rows: ", rows)
        print("\t Wells: ", wells)
        print("\t Tiles: ", tiles)

        if sort_wells:
            # update metadata to include destination for each tile
            metadata["dest"] = [
                os.path.join(self.outdirs["sorted_timepoints"], timepoint, f"{row}_{well}")
                for row, well, timepoint in zip(metadata.Row, metadata.Well, metadata.Timepoint, strict=False)
            ]
        else:
            metadata["dest"] = [
                os.path.join(self.outdirs["sorted_timepoints"], timepoint) for timepoint in metadata.Timepoint
            ]

        # unique directories for each tile
        unique_dirs = list(set(metadata.dest.to_list()))

        for _dir in unique_dirs:
            if not os.path.exists(_dir):
                os.makedirs(_dir)

        # copy/link the images to their new names
        self._copy_files(metadata=metadata)


class CombinedPhenixParser(PhenixParser):
    """
    Class to parse Phenix experiments where multiple experiments should be combined into one dataset.
    Usually this class is used if during image acquisition individual tiles were not imaged due to a focus failure.
    Instead of repeating the entire measurement, the missing tiles can be imaged in a separate experiment and then combined with the original dataset using this method.

    This class inherits from the PhenixParser class and extends it by adding the functionality to combine multiple experiments into one dataset.
    These individual experiments need to be placed together in the following structure:

        .. code-block:: none

            <experiment_name>/
            └── experiments_to_combine/
                ├── experiment_1/
                ├── experiment_2/
                ├── experiment_3/
                └── ...

        - *<experiment_name>* can be chosen freely.
        - *experiments_to_combine* is a folder containing all experiments that should be combined. This folder should be placed in the main experiment directory.
        - *experiment_n* always refers to the complete folder as generated by Harmony when exporting phenix data without any further modifications.

    The experiments will be combined in the order of their creation date and time.
    If two experiments contain images in the same position, the parser will keep the images from the first experiment.
    """

    directory_combined_measurements = "experiments_to_combine"

    def __init__(
        self,
        experiment_dir: str,
        flatfield_exported: bool = True,
        use_symlinks: bool = True,
        compress_rows: bool = False,
        compress_cols: bool = False,
        overwrite: bool = False,
    ) -> None:
        """
        Args:
            experiment_dir: The directory containing the exported phenxi experiment
            flatfield_exported: Whether flatfield corrected images were exported.
            use_symlinks: Whether to use symbolic links for parsed images.
            compress_rows: Whether to compress rows in the parsed images.
            compress_cols: Whether to compress columns in the parsed images.
            overwrite: Whether to overwrite existing files during the parsing process.
        """
        super().__init__(
            experiment_dir, flatfield_exported, use_symlinks, compress_rows, compress_cols, overwrite=overwrite
        )

        self.get_datasets_to_combine()

    def _get_xml_path(self):
        """Automatically get the XML files from all phenix experiments that should be combined."""
        # directory depends on if flatfield images were exported or not
        # these generated folder structures are hard coded during phenix export, do not change
        # get index file of the first phenix dir(this is our main experiment!)
        if self.flatfield_status:
            index_file_names = ["Index.xml", "Index.ref.xml"]
            for index_file_name in index_file_names:
                index_file = os.path.join(self.phenix_dirs[0], "Images", index_file_name)
                if os.path.isfile(index_file):
                    break
        else:
            index_file_names = ["Index.xml", "Index.idx.xml"]
            for index_file_name in index_file_names:
                index_file = os.path.join(self.phenix_dirs[0], "Images", index_file_name)
                if os.path.isfile(index_file):
                    break

        # perform sanity check if file exists else exit
        if not os.path.isfile(index_file):
            sys.exit(f"Can not find index file at path: {index_file}")

        return index_file

    def _get_input_dir(self) -> str:
        """Automatically get the subfolder where the exported image files are located."""

        # directory depends on if flatfield images were exported or not
        # these generated folder structures are hard coded during phenix export, do not change
        # for the combined exported the first experiment is always used (they should have the same exported XML file anyways for reading)
        if self.flatfield_status:
            input_dir = f"{self.phenix_dirs[0]}/Images/flex"
        else:
            input_dir = f"{self.phenix_dirs[0]}/Images"

        # perform sanity check if file exists else exit
        if not os.path.isdir(input_dir):
            sys.exit(f"Can not find directory containing images to parse: {input_dir}")

        return input_dir

    def get_datasets_to_combine(self) -> None:
        """Get all phenix experiments from subdirectories that should be combined."""
        input_path = f"{self.experiment_dir}/{self.directory_combined_measurements}"

        # get phenix directories that need to be comined together
        phenix_dirs = os.listdir(input_path)

        # only get the phenix dirs that match the pattern of a phenix experiment (ie they contain a time stamp)
        pattern = r"\d{4}-\d{2}-\d{2}T\d{2}_\d{2}_\d{2}"
        phenix_dirs = [x for x in phenix_dirs if re.search(pattern, x)]

        # Extract the date and time information from each file name
        dates_times = [
            re.search(pattern, file_name).group(0)
            for file_name in phenix_dirs
            if re.search(pattern, file_name)  # Ensure it matched before accessing .group()
        ]

        # Sort the file names based on the extracted date and time information
        sorted_phenix_dirs = [file_name for _, file_name in sorted(zip(dates_times, phenix_dirs, strict=False))]

        self.phenix_dirs = [f"{input_path}/{phenix_dir}" for phenix_dir in sorted_phenix_dirs]

    def _get_phenix_metadata(self) -> None:
        """Read combined metadata from all phenix experiments."""
        ###
        # read metadata from all experiments and merge into one file
        # note: if more than one image exists at a specific position then the first image aquired will be preserved based on the timestamps in the exported phenix measurement names
        ####

        # define under what path the actual exported images will be found
        # this is hard coded through phenix export script, do not change
        if self.flatfield_status:
            xml_path = "Images/Index.ref.xml"
            append_string = "Images/flex"
        else:
            append_string = "Images"
            xml_path = "Index.idx.xml"

        # read all metadata
        metadata = {}
        for phenix_dir in self.phenix_dirs:
            df = self._read_phenix_xml(f"{phenix_dir}/{xml_path}")
            df = df.set_index(["Row", "Well", "Zstack", "Timepoint", "X", "Y", "Channel"])
            df.loc[:, "source"] = f"{phenix_dir}/{append_string}"  # update source with the correct strings
            metadata[phenix_dir] = df

        # merge generated metadata files together (order of what is preserved is according to calcualted creation times above)
        for i, key in enumerate(metadata.keys()):
            if i == 0:
                metadata_merged = metadata[key]
            else:
                metadata_merged = metadata_merged.combine_first(metadata[key])

        metadata_merged = metadata_merged.reset_index()

        # return generated dataframe
        print("merged metadata generated from all passed phenix experiments.")
        return metadata_merged



"""
parse
====================================

Contains functions to parse imaging data aquired on an OperaPhenix or Operetta into usable formats for downstream pipelines.
"""

from ._parse_phenix import CombinedPhenixParser, PhenixParser

__all__ = ["PhenixParser", "CombinedPhenixParser"]



import os

import h5py
import matplotlib.pyplot as plt
import numpy as np
from skimage.color import label2rgb


def plot_image(array, size=(10, 10), save_name="", cmap="magma", return_fig=False, **kwargs):
    """
    Visualize and optionally save an input array as an image.

    This function displays a 2D array using matplotlib and can save
    the resulting image as a PNG file.

    Args:
        array (np.array): Input 2D numpy array to be plotted.
        size (tuple of int, optional): Figure size in inches, by default (10, 10).
        save_name (str, optional): Name of the output file, without extension. If not provided, image will not be saved, by default "".
        cmap (str, optional): Color map used to display the array, by default "magma".
        **kwargs: Additional keyword arguments to be passed to `ax.imshow`.

    Returns:
        None: The function will display the image but does not return any values.

    Example:
    >>> array = np.random.rand(10, 10)
    >>> plot_image(array, size=(5, 5))
    """

    fig = plt.figure(frameon=False)
    fig.set_size_inches(size)
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.imshow(array, cmap=cmap, **kwargs)

    if return_fig:
        return fig

    if save_name != "":
        plt.savefig(save_name + ".png")
        plt.show()
        plt.close()
    else:
        plt.show()
        plt.close()


def visualize_class(class_ids, seg_map, image, all_ids=None, return_fig=False, **kwargs):
    """
    Visualize specific classes in a segmentation map by highlighting them on top of a background image.

    This function takes in class IDs and a segmentation map, and creates an output visualization
    where the specified classes are highlighted on top of the provided background image.

    Args:
        class_ids (array-like): A list or array of integers representing the class IDs to be highlighted.
        seg_map (2D array-like): A 2D array representing the segmentation map, where each value corresponds to a class ID.
        background (2D/3D array-like): Background image (2D or 3D) on which the classes will be highlighted. Its size should match that of `seg_map`.
        *args: Any additional positional arguments that are passed to the underlying plotting functions.
        **kwargs: Any additional keyword arguments that are passed underlying plotting functions.

    Returns:
        None: The function will display the highlighted image but does not return any values.

    Example:
    >>> class_ids = [1, 2]
    >>> seg_map = np.array([[0, 1, 0], [1, 2, 1], [2, 0, 1]])
    >>> background = np.random.random((3, 3)) * 255
    >>> visualize_class(class_ids, seg_map, background)
    """

    # ensure the class ids are a list
    if not isinstance(class_ids, list):
        class_ids = list(class_ids)

    if all_ids is None:
        all_ids = set(np.unique(seg_map)) - {0}

    # get the ids to keep
    keep_ids = list(all_ids - set(class_ids))

    mask_discard = np.isin(seg_map, keep_ids)
    mask_keep = np.isin(seg_map, class_ids)

    # Set the values in the output map to 2 for the specified class IDs
    outmap = np.where(mask_discard, 2, seg_map)

    # Set the values in the output map to 1 for all class IDs other than the specified classes
    outmap = np.where(mask_keep, 1, outmap)

    vis_map = label2rgb(outmap, image=image, colors=["red", "blue"], alpha=0.4, bg_label=0)

    fig = plot_image(vis_map, return_fig=True, **kwargs)

    if return_fig:
        return fig


def plot_segmentation_mask(
    project,
    mask_channel=0,
    image_channel=0,
    selection=None,
    cmap_image="Greys_r",
    cmap_masks="prism",
    alpha=0.5,
    figsize=(10, 10),
):
    """
    Visualize the segmentation mask overlayed with a channel of the input image.

    Parameters
    ----------
    project : scPortrait.pipeline.project.Project
        instance of a scPortrait project.
    mask_channel : int, optional
        The index of the channel to use for the segmentation mask (default: 0).
    image_channel : int, optional
        The index of the channel to use for the image (default: 0).
    selection : tuple(slice, slice), optional
        The selection coordinates for a specific region of interest (default: None).
    cmap_image : str, optional
        The colormap to use for the input image (default: "Greys_r").
    cmap_masks : str, optional
        The colormap to use for the segmentation mask (default: "prism").
    alpha : float, optional
        The transparency level of the segmentation mask (default: 0.5).

    Returns
    -------
    fig : object
        The generated figure object.
    """
    # integer value indicating background (default value)
    background = 0

    segmentation_file = os.path.join(project.seg_directory, project.segmentation_f.DEFAULT_SEGMENTATION_FILE)

    with h5py.File(segmentation_file, "r") as hf:
        segmentation = hf.get(project.segmentation_f.DEFAULT_MASK_NAME)
        channels = hf.get(project.segmentation_f.DEFAULT_CHANNELS_NAME)

        if selection is None:
            segmentation = segmentation[mask_channel, :, :]
            image = channels[image_channel, :, :]
        else:
            segmentation = segmentation[mask_channel, selection[0], selection[1]]
            image = channels[image_channel, selection[0], selection[1]]

    # set background to np.nan so that its not visualized
    segmentation = np.where(segmentation == background, np.nan, segmentation)

    fig = plt.figure(figsize=figsize)
    plt.imshow(image, cmap=cmap_image)
    plt.imshow(segmentation, alpha=alpha, cmap=cmap_masks)
    plt.axis("off")
    return fig


def colorize(im: np.ndarray, color: tuple[int, ...] = (1, 0, 0), clip_percentile: float = 0.0):
    """
    Helper function to create an RGB image from a single-channel image using a
    specific color.

    Args:
        im (np.ndarray): A single-channel image.
        color (tuple[int, ...], optional): The color to use for the image. Defaults to (1, 0, 0).
        clip_percentile (float, optional): The percentile to clip the image at. Defaults to 0.0.

    Returns:
        np.ndarray: The colorized image.
    """
    # Check that we do just have a 2D image
    if im.ndim > 2 and im.shape[2] != 1:
        raise ValueError("This function expects a single-channel image!")

    # Rescale the image according to how we want to display it
    im_scaled = im.astype(np.float32) - np.percentile(im, clip_percentile)
    im_scaled = im_scaled / np.percentile(im_scaled, 100 - clip_percentile)
    im_scaled = np.clip(im_scaled, 0, 1)

    # Need to make sure we have a channels dimension for the multiplication to work
    im_scaled = np.atleast_3d(im_scaled)

    # Reshape the color (here, we assume channels last)
    color = np.asarray(color).reshape((1, 1, -1))
    return im_scaled * color


def generate_composite(images: np.ndarray, colors: list[tuple[int, ...]] = None, plot: bool = False):
    """Create a composite image from a multi-channel image for visualization.

    Args:
        images (np.ndarray): A multi-channel image to be combined.
        colors (list[tuple[int, ...]], optional): A list of colors to use for each channel. Defaults to None.
        plot (bool, optional): Whether to plot the composite image. Defaults to False.

    Returns:
        np.ndarray: The composite image.
    """
    if colors is None:
        colors = [(0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 0, 1)]
    colorized = []
    for image, color in zip(images, colors, strict=False):
        image = colorize(image, color, 0.0)
        colorized.append(image)

    if plot:
        for i in colorized:
            plt.figure()
            plt.imshow(i)

    image = colorized[0]
    for i in range(len(colorized) - 1):
        image += colorized[i + 1]

    return np.clip(image, 0, 1)



"""Utility functions for handling SpatialData objects and operations."""

from typing import TypeAlias, Union

import numpy as np
import pandas as pd
import psutil
import xarray
from dask.array import unique as DaskUnique
from spatialdata import SpatialData, get_centroids
from spatialdata._core.operations.transform import transform
from spatialdata.models import PointsModel, TableModel
from spatialdata.transformations.operations import get_transformation
from spatialdata.transformations.transformations import BaseTransformation

from scportrait.pipeline._utils.segmentation import numba_mask_centroid

# Type aliases
DataElement: TypeAlias = xarray.DataTree | xarray.DataArray
ChunkSize: TypeAlias = tuple[int, ...]
ChunkSizes: TypeAlias = list[tuple[int, ...]]


def check_memory(item: xarray.DataArray) -> bool:
    """Check if item can fit in available memory.

    Args:
        item: Array to check memory requirements for

    Returns:
        Whether item can fit in memory
    """
    array_size = item.nbytes
    available_memory = psutil.virtual_memory().available
    return array_size < available_memory


def generate_region_annotation_lookuptable(sdata: SpatialData) -> dict[str, list[tuple[str, TableModel]]]:
    """Generate a lookup table for the region annotation tables contained in a SpatialData object ordered according to the region they annotate.

    Args:
        sdata: SpatialData object to process

    Returns:
        Mapping of region names to list of (table_name, table) tuples
    """
    table_names = list(sdata.tables.keys())
    region_lookup: dict[str, list[tuple[str, TableModel]]] = {}

    for table_name in table_names:
        table = sdata.tables[table_name]
        region = table.uns["spatialdata_attrs"]["region"]

        if region not in region_lookup:
            region_lookup[region] = [(table_name, table)]
        else:
            region_lookup[region].append((table_name, table))

    return region_lookup


def remap_region_annotation_table(table: TableModel, region_name: str) -> TableModel:
    """Remap region annotation table to new region name.

    Args:
        table: Region annotation table to remap
        region_name: New region name

    Returns:
        Remapped table
    """
    instance_key = table.uns["spatialdata_attrs"]["instance_key"]
    new_instance_key = "cell_id"  # default value for scportrait runs

    table = table.copy()
    table.obs["region"] = region_name
    table.obs["region"] = table.obs["region"].astype("category")
    table.obs.rename(columns={instance_key: new_instance_key}, inplace=True)

    if "spatialdata_attrs" in table.uns:
        del table.uns["spatialdata_attrs"]  # remove the spatialdata attributes so that the table can be re-written

    return TableModel.parse(table, region_key="region", region=region_name, instance_key="cell_id")


def get_chunk_size(element: DataElement) -> ChunkSize | ChunkSizes:
    """Get chunk size of image data.

    Args:
        element: Element to get chunk size from

    Returns:
        Chunk size(s) of the image data

    Raises:
        ValueError: If element type is not supported
    """
    if isinstance(element, xarray.DataArray):
        if len(element.shape) == 2:
            y, x = element.chunksizes.values()
            y = y[0] if isinstance(y, tuple | list) or len(y) > 1 else y
            x = x[0] if isinstance(x, tuple | list) or len(x) > 1 else x
            return (int(y), int(x))
        elif len(element.shape) == 3:
            c, y, x = element.chunksizes.values()
            c = c[0] if isinstance(c, tuple | list) or len(c) > 1 else c
            y = y[0] if isinstance(y, tuple | list) or len(y) > 1 else y
            x = x[0] if isinstance(x, tuple | list) or len(x) > 1 else x
            return (int(c), int(y), int(x))

    elif isinstance(element, xarray.DataTree):
        chunk_sizes: ChunkSizes = []
        for scale in element:
            if len(element[scale]["image"].shape) == 2:
                y, x = element[scale].chunksizes.values()
                y = y[0] if len(y) > 1 else y
                x = x[0] if len(x) > 1 else x
                chunk_sizes.append((int(y), int(x)))
            elif len(element[scale]["image"].shape) == 3:
                c, y, x = element[scale].chunksizes.values()
                c = c[0] if len(c) > 1 else c
                y = y[0] if len(y) > 1 else y
                x = x[0] if len(x) > 1 else x
                chunk_sizes.append((int(c), int(y), int(x)))
        return chunk_sizes

    raise ValueError(f"Element must be DataTree or DataArray, found {type(element)}")


def rechunk_image(element: DataElement, chunk_size: ChunkSize) -> DataElement:
    """Rechunk image data to desired chunk size.

    Args:
        element: Element to rechunk
        chunk_size: Desired chunk dimensions

    Returns:
        Rechunked element

    Raises:
        ValueError: If element type is not supported
    """
    if isinstance(element, xarray.DataArray):
        element.data = element.data.rechunk(chunk_size)
        return element
    elif isinstance(element, xarray.DataTree):
        for scale in element:
            element[scale]["image"].data = element[scale]["image"].data.rechunk(chunk_size)
        return element
    raise ValueError(f"Element must be DataTree or DataArray, found {type(element)}")


def make_centers_object(
    centers: np.ndarray, ids: list[int], transformation: BaseTransformation, coordinate_system: str = "global"
) -> PointsModel:
    """Create PointsModel from centers and IDs.

    Args:
        centers: Array of center coordinates
        ids: List of point IDs
        transformation: Transformation to apply
        coordinate_system: Coordinate system name

    Returns:
        Points model containing the centers
    """
    coordinates = pd.DataFrame(centers, columns=["y", "x"], index=ids)
    centroids = PointsModel.parse(coordinates, transformations={coordinate_system: transformation})
    return transform(centroids, to_coordinate_system=coordinate_system)


def calculate_centroids(mask: xarray.DataArray, coordinate_system: str = "global") -> PointsModel:
    """Calculate centroids of labeled regions.

    Args:
        mask: Labeled mask
        coordinate_system: Coordinate system name

    Returns:
        Points model containing centroids
    """
    transform = get_transformation(mask, coordinate_system)

    if check_memory(mask):
        centers, _, _ids = numba_mask_centroid(mask.values)
        return make_centers_object(centers, _ids, transform, coordinate_system)

    print("Array larger than memory, using dask-delayed calculation.")
    return get_centroids(mask, coordinate_system)


def get_unique_cell_ids(data: xarray.DataArray, remove_background: bool = True) -> set[int]:
    """Get unique cell IDs from segmentation mask.

    Args:
        data: Segmentation mask
        remove_background: Whether to remove background (0) label

    Returns:
        Set of unique cell IDs
    """
    cell_ids = set(DaskUnique(data.data).compute())
    if remove_background:
        cell_ids = cell_ids - {0}
    return cell_ids



from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F


### VGG Model Architecture
class VGGBase(nn.Module):
    """
    Base implementation of VGG Model Architecture. Can be implemented with varying number of
    convolutional neural layers and fully connected layers.

    Args:
        nn (Module): Pytorch Module

    Returns:
        nn.Module: Pytorch Module
    """

    cfgs = {
        "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
        "B": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M", 512, "M"],
        "D": [64, 64, "M", 128, 128, "M", 256, 256, 256, "M", 512, 512, 512, "M", 512, 512, 512, "M"],
        "E": [64, 64, "M", 128, 128, "M", 256, 256, 256, 256, "M", 512, 512, 512, 512, "M", 512, 512, 512, 512, "M"],
    }

    cfgs_MLP = {
        "A": [
            2048,
            "M",
            1024,
            "M",
        ],
        "B": [
            1024,
            "M",
            512,
            "M",
            256,
            "M",
        ],
    }

    def make_layers(self, cfg: list, in_channels: int, batch_norm: bool = True):
        """
        Create sequential models layers according to the chosen configuration provided in
        cfg with optional batch normalization for the CNN.

        Args:
            cfg (list): A list of integers and “M” representing the specific
            VGG architecture of the CNN
            in_channels (int): Number of input channels
            batch_norm (bool, optional): Whether to use batch normalization. Defaults to True.

        Returns:
            nn.Sequential: A sequential model representing the VGG architecture.
        """
        layers = []
        i = 0
        for v in cfg:
            if v == "M":
                layers += [(f"maxpool{i}", nn.MaxPool2d(kernel_size=2, stride=2))]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                if batch_norm:
                    layers += [
                        (f"conv{i}", conv2d),
                        (f"batchnorm{i}", nn.BatchNorm2d(v)),
                        (f"relu{i}", nn.ReLU(inplace=True)),
                    ]
                else:
                    layers += [(f"conv{i}", conv2d), (f"relu{i}", nn.ReLU(inplace=True))]
                in_channels = v
                i += 1
        return nn.Sequential(OrderedDict(layers))

    def make_layers_MLP(self, cfg_MLP: list, cfg: list, image_size_factor: int = 2):
        """
        Create sequential models layers according to the chosen configuration provided in
        cfg for the MLP.

        Args:
            cfg_MLP (list): A list of integers and “M” representing the specific
            VGG architecture of the MLP
            cfg (list): A list of integers and “M” representing the specific
            VGG architecture of the CNN
            image_size_factor (int, optional): VGG model as implemented expects input images of 128x128 pixels.
            Defaults to 2. For example, if the input image is 256x256, the image_size_factor should be 4.

        Returns:
            nn.Sequential: A sequential model representing the MLP architecture.
        """
        # get output feature size of CNN with chosen configuration
        in_features = int(cfg[-2]) * image_size_factor * image_size_factor

        layers = []
        i = 0
        for out_features in cfg_MLP:
            if out_features == "M":
                layers += [(f"MLP_relu{i}", nn.ReLU(True)), (f"MLP_dropout{i}", nn.Dropout())]
                i += 1
            else:
                linear = (f"MLP_linear{i}", nn.Linear(in_features, out_features))
                layers += [linear]
                in_features = out_features

        linear = (f"MLP_linear{i}_classes", nn.Linear(in_features, self.num_classes))
        layers += [linear]

        return nn.Sequential(OrderedDict(layers))

    def forward(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        x = self.softmax(x)
        return x

    def encoder(self, x):
        x = self.norm(x)
        x = self.features(x)
        return torch.flatten(x, 1)


class VGG1(VGGBase):
    """
    Instance of VGGBase with the model architecture 1.
    """

    def __init__(
        self,
        cfg="B",  # default configuration
        cfg_MLP="A",
        dimensions=196,
        in_channels=1,
        num_classes=2,
        image_size_factor=2,
    ):
        super().__init__()

        # save num_classes for use in making MLP head
        self.num_classes = num_classes
        self.image_size_factor = image_size_factor

        self.norm = nn.BatchNorm2d(in_channels)
        self.softmax = nn.LogSoftmax(dim=1)

        self.features = self.make_layers(self.cfgs[cfg], in_channels)
        self.classifier = self.make_layers_MLP(
            self.cfgs_MLP[cfg_MLP], self.cfgs[cfg], image_size_factor=image_size_factor
        )

    def vgg(self, cfg, in_channels, **kwargs):
        model = VGG1(self.make_layers(self.cfgs[cfg], in_channels), **kwargs)
        return model


class VGG2(VGGBase):
    """
    Instance of VGGBase with the model architecture 2.
    """

    def __init__(
        self,
        cfg="B",
        cfg_MLP="B",
        dimensions=196,
        in_channels=1,
        num_classes=2,
        image_size_factor=2,
    ):
        super().__init__()

        # save num_classes for use in making MLP head
        self.num_classes = num_classes
        self.image_size_factor = image_size_factor

        self.norm = nn.BatchNorm2d(in_channels)
        self.softmax = nn.LogSoftmax(dim=1)

        self.features = self.make_layers(self.cfgs[cfg], in_channels)
        self.classifier = self.make_layers_MLP(
            self.cfgs_MLP[cfg_MLP], self.cfgs[cfg], image_size_factor=image_size_factor
        )

    def vgg(self, cfg, in_channels, **kwargs):
        model = VGG2(self.make_layers(self.cfgs[cfg], in_channels), **kwargs)
        return model


class VGG2_regression(VGGBase):
    """
    Instance of VGGBase with regression output.
    """

    def __init__(
        self,
        cfg="B",
        cfg_MLP="A",
        in_channels=1,
        num_classes=1,
        image_size_factor=2,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.image_size_factor = image_size_factor
        self.norm = nn.BatchNorm2d(in_channels)

        self.features = self.make_layers(self.cfgs[cfg], in_channels)
        self.classifier = self.make_layers_MLP(
            self.cfgs_MLP[cfg_MLP], self.cfgs[cfg], image_size_factor=image_size_factor
        )

    def vgg(self, cfg, in_channels, **kwargs):
        model = VGG2_regression(self.make_layers(self.cfgs[cfg], in_channels), **kwargs)
        return model

    def forward(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x


### CAE Model Architecture


class CAEBase(nn.Module):
    """
    Base implementation of CAE Model Architecture.
    """

    en_cfgs = {
        "A": [16, "M", 32, "M", 64, "M", 128, "M", 256, "M", 512, "M"],
        "A_deep": [32, "M", 64, "M", 128, "M", 256, "M", 512, 512, "M", 512, 512, "M"],
        "B": [16, "M", 32, "M", 64, "M", 128, "M", 256, "M", 256, "M"],
        "C": [16, "M", 32, "M", 64, "M", 128, "M", 256, "M", 512, "M", 512, "M"],
        "D": [16, "M", 32, "M", 64, "M", 128, "M", 256, "M", 256, "M", 256, "M"],
        "E": [16, "M", 32, "M", 64, "M", 128, "M", 256, "M", 256, "M", 128, "M"],
    }

    de_cfgs = {
        "A": [512, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16, "U", 16],
        "A_deep": [512, 512, "U", 512, 512, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16],
        "B": [256, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16, "U", 16],
        "C": [512, "U", 512, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16, "U", 16],
        "D": [256, "U", 256, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16, "U", 16],
        "E": [128, "U", 256, "U", 256, "U", 128, "U", 64, "U", 32, "U", 16, "U", 16],
    }

    latent_dim = {
        "A": 2048,
        "A_deep": 2048,
        "B": 1024,
        "C": 512,
        "D": 256,
        "E": 128,
    }

    initial_decoder_depth = {
        "A": 512,
        "A_deep": 512,
        "B": 256,
        "C": 512,
        "D": 256,
        "E": 128,
    }

    def __init__(
        self,
        cfg="B",
        in_channels=5,
        out_channels=5,
    ):
        super().__init__()

        self.norm = nn.BatchNorm2d(in_channels)

        self.encoder = self.make_encoder(self.en_cfgs[cfg], in_channels)
        self.decoder = self.make_decoder(self.de_cfgs[cfg], self.initial_decoder_depth[cfg], out_channels)

    def latent(self, x):
        x = self.norm(x)
        x = self.encoder(x)
        x = torch.flatten(x, 1)
        return x

    def forward(self, x):
        x = self.norm(x)
        x = self.encoder(x)

        shape2d = x.shape
        x = torch.flatten(x, 1)

        x = torch.reshape(x, shape2d)

        x = self.decoder(x)
        x = torch.sigmoid(x)

        return x

    def make_encoder(self, cfg, in_channels, batch_norm=True):
        layers = []
        for v in cfg:
            if v == "M":
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                if batch_norm:
                    layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
                else:
                    layers += [conv2d, nn.ReLU(inplace=True)]
                in_channels = v
        return nn.Sequential(*layers)

    def make_decoder(self, cfg, in_channels, out_channels):
        layers = []

        for v in cfg:
            if v == "U":
                layers += [nn.Upsample(scale_factor=2)]
            else:
                tconv2d = nn.ConvTranspose2d(in_channels, v, 3, padding=1, stride=1)
                layers += [tconv2d, nn.ReLU(inplace=True)]
                in_channels = v

        layers += [nn.ConvTranspose2d(in_channels, out_channels, 3, padding=1, stride=1)]

        return nn.Sequential(*layers)


### VAE Model Architecture


class VAEBase(nn.Module):
    """
    Base implementation of VAE Model Architecture.
    """

    def __init__(self, in_channels, out_channels, latent_dim, hidden_dims=None, **kwargs):
        super().__init__()

        self.latent_dim = latent_dim

        modules = []
        if hidden_dims is None:
            hidden_dims = [32, 64, 128, 256, 512, 512, 512]

        # Build Encoder
        for h_dim in hidden_dims:
            modules.append(
                nn.Sequential(
                    nn.Conv2d(in_channels, out_channels=h_dim, kernel_size=3, stride=2, padding=1),
                    nn.BatchNorm2d(h_dim),
                    nn.LeakyReLU(),
                )
            )
            in_channels = h_dim

        self.encoder = nn.Sequential(*modules)
        self.fc_mu = nn.Linear(hidden_dims[-1], latent_dim)
        self.fc_var = nn.Linear(hidden_dims[-1], latent_dim)

        # Build Decoder
        modules = []

        self.decoder_input = nn.Linear(latent_dim, hidden_dims[-1])

        hidden_dims.reverse()

        for i in range(len(hidden_dims) - 1):
            modules.append(
                nn.Sequential(
                    nn.ConvTranspose2d(
                        hidden_dims[i], hidden_dims[i + 1], kernel_size=3, stride=2, padding=1, output_padding=1
                    ),
                    nn.BatchNorm2d(hidden_dims[i + 1]),
                    nn.LeakyReLU(),
                )
            )

        self.decoder = nn.Sequential(*modules)

        self.final_layer = nn.Sequential(
            nn.ConvTranspose2d(hidden_dims[-1], hidden_dims[-1], kernel_size=3, stride=2, padding=1, output_padding=1),
            nn.BatchNorm2d(hidden_dims[-1]),
            nn.LeakyReLU(),
            nn.Conv2d(hidden_dims[-1], out_channels=out_channels, kernel_size=3, padding=1),
            nn.Sigmoid(),
        )

    def encode(self, input):
        """
        Encodes the input by passing through the encoder network
        and returns the latent codes.

        Args:
            input (torch.Tensor): Input tensor to the encoder with dimensions [N x C x H x W].

        Returns:
            list: A list containing mean and log variance components of the latent Gaussian distribution.
        """
        result = self.encoder(input)

        result = torch.flatten(result, start_dim=1)

        # Split the result into mu and var components
        # of the latent Gaussian distribution
        mu = self.fc_mu(result)
        log_var = self.fc_var(result)

        return [mu, log_var]

    def decode(self, z):
        """
        Maps the given latent codes onto the image space.

        Args:
            z (torch.Tensor): Latent code tensor with dimensions [B x D].

        Returns:
            torch.Tensor: Reconstructed tensor with dimensions [B x C x H x W].
        """
        result = self.decoder_input(z)
        result = result.view(-1, 512, 1, 1)
        result = self.decoder(result)
        result = self.final_layer(result)
        return result

    def reparameterize(self, mu, logvar):
        """
        Reparameterization trick to sample from N(mu, var) from N(0,1).

        Args:
            mu (torch.Tensor): Mean of the latent Gaussian with dimensions [B x D].
            logvar (torch.Tensor): Log variance of the latent Gaussian with dimensions [B x D].

        Returns:
            torch.Tensor: Reparameterized tensor with dimensions [B x D].
        """
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps * std + mu

    def forward(self, input, **kwargs):
        """
        Performs the forward pass of the VAE model.

        The input passes through the encoder to obtain the mean and log variance.
        Then, the reparameterization trick is applied to generate the latent vector z.
        Finally, z is passed through the decoder to generate the reconstructed input.

        Args:
            input (torch.Tensor): The input tensor to be passed through the VAE model.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            list: A list containing the reconstructed input, mean and log variance.
        """
        mu, log_var = self.encode(input)
        z = self.reparameterize(mu, log_var)
        return [self.decode(z), mu, log_var]

    def loss_function(self, target, output, *args, **kwargs):
        r"""
        Computes the VAE loss function.

        .. math::
            KL(N(\mu, \sigma), N(0, 1)) =  \log \frac{1}{\sigma} + \frac{\sigma^2 + \mu^2}{2} - \frac{1}{2}

        Args:
            target (torch.Tensor): The target tensor.
            output (tuple): A tuple containing the reconstructed tensor, mean, and log variance.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: A dictionary containing the total loss, reconstruction loss, and KL divergence loss.
        """
        recons = output[0]
        mu = output[1]
        log_var = output[2]

        kld_weight = kwargs["M_N"]  # Account for the minibatch samples from the dataset
        recons_loss = F.mse_loss(recons, target)

        kld_loss = torch.mean(-0.5 * torch.sum(1 + log_var - mu**2 - log_var.exp(), dim=1), dim=0)

        loss = recons_loss + kld_weight * kld_loss
        return {"loss": loss, "Reconstruction_Loss": recons_loss, "KLD": -kld_loss}


#### DEPRECATED FUNCTIONS FOR BACKWARD COMPATABILITY


class _VGG1(nn.Module):
    cfgs = {
        "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
        "B": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M", 512, "M"],
        "D": [64, 64, "M", 128, 128, "M", 256, 256, 256, "M", 512, 512, 512, "M", 512, 512, 512, "M"],
        "E": [64, 64, "M", 128, 128, "M", 256, 256, 256, 256, "M", 512, 512, 512, 512, "M", 512, 512, 512, 512, "M"],
    }

    def __init__(
        self,
        cfg="B",
        dimensions=196,
        in_channels=5,
        num_classes=2,
    ):
        super().__init__()

        self.norm = nn.BatchNorm2d(in_channels)

        # self.avgpool = nn.AdaptiveAvgPool2d((4, 4))

        self.softmax = nn.LogSoftmax(dim=1)

        self.features = self.make_layers(self.cfgs[cfg], in_channels)

        self.classifier_1 = nn.Sequential(
            nn.Linear(512 * 2 * 2, 2048),
        )

        self.classifier_2 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(2048, 1024),
        )

        self.classifier_3 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(1024, num_classes),
        )

    def forward(self, x):
        x = self.norm(x)
        x = self.features(x)
        # x = self.avgpool(x)

        x = torch.flatten(x, 1)

        x = self.classifier_1(x)
        x = self.classifier_2(x)
        x = self.classifier_3(x)

        x = self.softmax(x)
        return x

    def encoder(self, x):
        x = self.norm(x)
        x = self.features(x)
        return torch.flatten(x, 1)

    def encoder_c1(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        return x

    def encoder_c2(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        x = self.classifier_2(x)
        return x

    def encoder_c3(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        x = self.classifier_2(x)
        x = self.classifier_3(x)
        return x

    def make_layers(self, cfg, in_channels, batch_norm=True):
        layers = []
        for v in cfg:
            if v == "M":
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                if batch_norm:
                    layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
                else:
                    layers += [conv2d, nn.ReLU(inplace=True)]
                in_channels = v
        return nn.Sequential(*layers)

    def vgg(self, cfg, in_channels, **kwargs):
        model = _VGG1(self.make_layers(self.cfgs[cfg], in_channels), **kwargs)
        return model


class _VGG2(nn.Module):
    cfgs = {
        "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
        "B": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M", 512, "M"],
        "D": [64, 64, "M", 128, 128, "M", 256, 256, 256, "M", 512, 512, 512, "M", 512, 512, 512, "M"],
        "E": [64, 64, "M", 128, 128, "M", 256, 256, 256, 256, "M", 512, 512, 512, 512, "M", 512, 512, 512, 512, "M"],
    }

    def __init__(
        self,
        cfg="B",
        dimensions=196,
        in_channels=5,
        num_classes=2,
    ):
        super().__init__()

        self.norm = nn.BatchNorm2d(in_channels)
        # self.avgpool = nn.AdaptiveAvgPool2d((4, 4))

        self.softmax = nn.LogSoftmax(dim=1)

        self.features = self.make_layers(self.cfgs[cfg], in_channels)

        self.classifier_1 = nn.Sequential(
            nn.Linear(512 * 2 * 2, 1024),
        )

        self.classifier_2 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(1024, 512),
        )

        self.classifier_3 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(512, 256),
        )

        self.classifier_4 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(256, num_classes),
        )

    def forward(self, x):
        x = self.norm(x)
        x = self.features(x)
        # x = self.avgpool(x)

        x = torch.flatten(x, 1)

        x = self.classifier_1(x)
        x = self.classifier_2(x)
        x = self.classifier_3(x)
        x = self.classifier_4(x)

        x = self.softmax(x)
        return x

    def encoder(self, x):
        x = self.norm(x)
        x = self.features(x)
        return torch.flatten(x, 1)

    def encoder_c1(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        return x

    def encoder_c2(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        x = self.classifier_2(x)
        return x

    def encoder_c3(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        x = self.classifier_2(x)
        x = self.classifier_3(x)
        return x

    def encoder_c4(self, x):
        x = self.norm(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier_1(x)
        x = self.classifier_2(x)
        x = self.classifier_3(x)
        x = self.classifier_4(x)
        return x

    def make_layers(self, cfg, in_channels, batch_norm=True):
        layers = []
        for v in cfg:
            if v == "M":
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                if batch_norm:
                    layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
                else:
                    layers += [conv2d, nn.ReLU(inplace=True)]
                in_channels = v
        return nn.Sequential(*layers)

    def vgg(self, cfg, in_channels, **kwargs):
        model = _VGG1(self.make_layers(self.cfgs[cfg], in_channels), **kwargs)
        return model



import numpy as np
import xarray as xr
from numba import jit
from skimage.exposure import rescale_intensity


def rescale_image(
    image: np.ndarray,
    rescale_range: tuple[int, int],
    outrange: tuple[float, float] = (0, 1),
    dtype="uint16",
    cutoff_threshold: int = None,
    return_float: bool = False,
):
    # convert to float for better percentile calculation
    img = image.astype(float)

    # define factor to rescale image to uints
    if dtype == "uint16":
        factor = 65535
    elif dtype == "uint8":
        factor = 255

    # get cutoff values
    cutoff1, cutoff2 = rescale_range

    # if cutoff_threshold is given, set all values above the threshold to 1 but do not consider them for percentile calculation
    if cutoff_threshold is not None:
        if img.max() > cutoff_threshold:
            _img = img.copy()
            values = _img[_img < (cutoff_threshold / factor)]
    else:
        values = img

    # calculate percentiles
    p1 = np.percentile(values, cutoff1)
    p99 = np.percentile(values, cutoff2)
    del values

    if return_float:
        return rescale_intensity(img, (p1, p99), outrange)
    else:
        return (rescale_intensity(img, (p1, p99), outrange) * factor).astype(dtype)


def _normalize_image(im: np.ndarray, lower_value: int, upper_value: int) -> np.ndarray:
    """
    Normalize an input 2D image (np.array) based on input values.

    Args:
        im: Numpy array of shape (height, width).
        lower_value: Lower image value used for normalization, all lower values will be clipped to 0.
        upper_value: Upper image value used for normalization, all higher values will be clipped to 1.

    Returns:
        out_im (np.array): Normalized Numpy array. Values are clipped to [0, 1].

    Example:
    >>> img = np.random.rand(4, 4)
    >>> norm_img = _normalize_image(img, 200, 15000)
    """

    # Compute inter-percentile range (IPR)
    IPR = upper_value - lower_value

    # Normalize the image
    out_im = im - lower_value

    if IPR != 0:  # add check to make sure IPR is not 0
        out_im = out_im / IPR

    # Clip values outside [0, 1]
    out_im = np.clip(out_im, 0, 1)

    return out_im


def _percentile_norm(im: np.ndarray, lower_percentile: float, upper_percentile: float) -> np.ndarray:
    """
    Normalize an input 2D image (np.array) based on the defined percentiles.

    This is an helper function used in the percentile_normalization function.

    Args:
        im: Numpy array of shape (height, width).
        lower_percentile: Lower percentile used for normalization, all lower values will be clipped to 0.
        upper_percentile: Upper percentile used for normalization, all higher values will be clipped to 1.

    Returns:
        Normalized numpy array.

    Example:
    >>> img = np.random.rand(4, 4)
    >>> norm_img = _percentile_norm(img, 0.001, 0.999)
    """

    # Calculate the quantiles
    lower_value = np.quantile(np.ravel(im), lower_percentile)
    upper_value = np.quantile(np.ravel(im), upper_percentile)

    out_im = _normalize_image(im, lower_value, upper_value)

    return out_im


def percentile_normalization(
    im: np.ndarray, lower_percentile: float = 0.001, upper_percentile: float = 0.999, return_copy: bool = True
) -> np.ndarray:
    """
    Normalize an input image channel-wise based on defined percentiles.

    The percentiles will be calculated, and the image will be normalized to [0, 1]
    based on the lower and upper percentile.

    Args:
        im: Numpy array of shape (height, width) or (channels, height, width).
        lower_percentile: Lower percentile used for normalization, all lower values will be clipped to 0. Defaults to 0.001.
        upper_percentile: Upper percentile used for normalization, all higher values will be clipped to 1. Defaults to 0.999.

    Returns:
        im (np.array): Normalized Numpy array with dtype == float

    Example:
    >>> img = np.random.rand(3, 4, 4)  # (channels, height, width)
    >>> norm_img = percentile_normalization(img, 0.001, 0.999)
    """
    # if false the input image is directly updated
    if return_copy:
        im = im.copy()  # ensure we are working on a copy not the original data

    # ensure that the dtype is converted to a float before running this function
    if not isinstance(im.dtype, float):
        im = im.astype(np.float32)

    # chek if data is passed as (height, width) or (channels, height, width)
    if len(im.shape) == 2:
        im = _percentile_norm(im, lower_percentile, upper_percentile)

    elif len(im.shape) == 3:
        for i, _channel in enumerate(im):
            im[i] = _percentile_norm(im[i], lower_percentile, upper_percentile)

    else:
        raise ValueError("Input dimensions should be (height, width) or (channels, height, width).")

    return im


def downsample_img(img: np.ndarray, N: int = 2, return_dtype=np.uint16) -> np.ndarray:
    """
    Function to downsample an image in shape CXY equivalent to NxN binning using the mean between pixels.
    Takes a numpy array image as input and returns a numpy array as uint16.

    Args:
        img : image to downsample
        N : number of pixels that should be binned together using mean between pixels

    Returns:
        downsampled array
    """
    downsampled = xr.DataArray(img, dims=["c", "x", "y"]).coarsen(c=1, x=N, y=N, boundary="exact").mean()
    downsampled = (downsampled / downsampled.max() * np.iinfo(return_dtype).max).astype(return_dtype)
    return np.array(downsampled)


def downsample_img_pxs(img: np.ndarray, N: int = 2) -> np.ndarray:
    """
    Function to downsample an image in shape CXY equivalent to taking every N'th pixel from each dimension.
    Channels are preserved as is. This does not use any interpolation and is therefore faster than the mean
    method but is less precise.

    Args:
        img : image to downsample
        N : the nth pixel to take from each dimension

    Returns:
        downsampled array

    Example:
    >>> img = np.random.rand(10, 10)
    >>> downsampled_img = downsample_img
    """

    downsampled = img[:, 0:-1:N, 0:-1:N]
    return downsampled


def downsample_img_padding(img: np.ndarray, N: int = 2) -> np.ndarray:
    """
    Downsample image by a factor of N by taking every Nth pixel.
    Before downsampling this function will pad the image to ensure its compatible with the selected kernel size.

    Args:
        img: image to be downsampled
        N: factor by which the image should be downsampled

    Returns:
        downsampled image

    Example:
    >>> img = np.random.rand(11, 11)
    >>> downsampled_img = downsample_img_padding(img)

    """

    # check if N fits perfectly into image shape if not calculate how much we need to pad
    if len(img.shape) == 3:
        _, x, y = img.shape
    elif len(img.shape) == 2:
        x, y = img.shape

    if x % N == 0:
        pad_x = (0, 0)
    else:
        pad_x = (0, N - x % N)

    if y % N == 0:
        pad_y = (0, 0)
    else:
        pad_y = (0, N - y % N)

    print(f"Performing image padding to ensure that image is compatible with selected downsample kernel size of {N}.")

    # perform image padding to ensure that image is compatible with downsample kernel size
    img = np.pad(img, ((0, 0), pad_x, pad_y))

    print(f"Downsampling image by a factor of {N}x{N}")

    # actually perform downsampling
    img = downsample_img(img, N=N)

    return img


@jit(nopython=True, parallel=True)  # Set "nopython" mode for best performance, equivalent to @njit
def rolling_window_mean(array: np.ndarray, size: int, scaling: bool = False) -> np.ndarray:
    """
    Compute rolling window mean and normalize the input 2D array.

    The function takes an input 2D array and a window size. It calculates the mean
    within a rolling window of provided size and removes the mean from each element
    in the window. The modified array is returned. If scaling is set to True,
    the chunk is normalized by dividing by its standard deviation.
    Function is numba optimized.

    Args:
        array: Input 2D numpy array.
        size: Size of the rolling window.
        scaling: If True, normalizes the chunk by dividing it by standard deviation, by default False.

    Returns:
        Processed 2D numpy array.

    Example:
    >>> array = np.random.rand(10, 10)
    >>> rolling_array = rolling_window_mean(array, size=5, scaling=False)
    """

    overlap = 0
    lengthy, lengthx = array.shape
    delta = size - overlap

    # calculate the number of steps given the window size
    ysteps = lengthy // delta
    xsteps = lengthx // delta

    x = 0
    y = 0

    # Iterate through the array with the given window size
    for i in range(ysteps):
        for j in range(xsteps):
            y = i * delta
            x = j * delta

            yd = min(y + size, lengthy)
            xd = min(x + size, lengthx)

            # Extract the current window (chunk) and calculate statistics
            chunk = array[y:yd, x:xd]
            std = np.std(chunk.flatten())
            max = np.max(chunk.flatten())

            # Scale the chunk, if scaling is True
            if scaling:
                chunk = chunk / std

            # Compute the mean and remove it from the chunk
            mean = np.median(chunk.flatten())

            if max > 0:
                chunk = chunk - mean
                chunk = np.where(chunk < 0, 0, chunk)

            array[y:yd, x:xd] = chunk

    if scaling:
        array = array / np.max(array.flatten())
    return array


def MinMax(inarr: np.ndarray) -> np.ndarray:
    """
    Perform Min-Max normalization on the input array.
    The function scales the input array values to the range [0, 1] using Min-Max
    normalization. If the range of the input array (i.e., max - min) is zero,
    the original array is returned unchanged.

    Args:
        inarr: Input 2D numpy array.

    Returns:
        Min-Max normalized numpy array.

    Example:
    >>> array = np.random.rand(10, 10)
    >>> normalized_array = MinMax(array)
    """

    if np.max(inarr) - np.min(inarr) > 0:
        return (inarr - np.min(inarr)) / (np.max(inarr) - np.min(inarr))
    else:
        return inarr
